"""Unit tests for :mod:`~tmlt.core.transformations.spark_transformations.join`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2022

from typing import List, Optional, Union, cast

import pandas as pd
from parameterized import parameterized
from pyspark.sql import types as st

from tmlt.core.domains.collections import DictDomain
from tmlt.core.domains.spark_domains import (
    SparkDataFrameDomain,
    SparkFloatColumnDescriptor,
    SparkIntegerColumnDescriptor,
    SparkStringColumnDescriptor,
)
from tmlt.core.metrics import (
    AddRemoveKeys,
    DictMetric,
    HammingDistance,
    IfGroupedBy,
    RootSumOfSquared,
    SumOf,
    SymmetricDifference,
)
from tmlt.core.transformations.spark_transformations.join import (
    PrivateJoin,
    PrivateJoinOnKey,
    PublicJoin,
    TruncationStrategy,
)
from tmlt.core.utils.exact_number import ExactNumberInput
from tmlt.core.utils.testing import (
    PySparkTest,
    TestComponent,
    assert_property_immutability,
    get_all_props,
)


class TestPublicJoin(TestComponent):
    """Tests for class PublicJoin.

    Tests :class:`~tmlt.core.transformations.spark_transformations.join.PublicJoin`.
    """

    def setUp(self):
        """Test setup."""
        self.input_domain = SparkDataFrameDomain(
            {"A": SparkFloatColumnDescriptor(), "B": SparkStringColumnDescriptor()}
        )
        self.public_df = self.spark.createDataFrame(
            [("X", 10.0), ("X", 11.0)],
            schema=st.StructType(
                [
                    st.StructField("B", st.StringType(), nullable=False),
                    st.StructField("C", st.DoubleType(), nullable=False),
                ]
            ),
        )
        self.private_df = self.spark.createDataFrame(
            [(1.2, "X")],
            schema=st.StructType(
                [
                    st.StructField("A", st.DoubleType(), nullable=False),
                    st.StructField("B", st.StringType(), nullable=False),
                ]
            ),
        )

    def test_constructor_mutable_arguments(self):
        """Tests that mutable constructor arguments are copied."""
        join_cols = ["B"]
        transformation = PublicJoin(
            input_domain=self.input_domain,
            metric=SymmetricDifference(),
            public_df=self.public_df,
            join_cols=join_cols,
        )
        join_cols.append("C")
        self.assertListEqual(transformation.join_cols, ["B"])

    @parameterized.expand(get_all_props(PublicJoin))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        transformation = PublicJoin(
            input_domain=self.input_domain,
            metric=SymmetricDifference(),
            public_df=self.public_df,
            join_cols=["B"],
        )
        assert_property_immutability(transformation, prop_name)

    def test_properties(self):
        """PublicJoin's properties have the expected values."""
        transformation = PublicJoin(
            input_domain=self.input_domain,
            metric=SymmetricDifference(),
            public_df=self.public_df,
            join_cols=["B"],
        )
        self.assertEqual(transformation.input_domain, self.input_domain)
        self.assertEqual(transformation.input_metric, SymmetricDifference())
        self.assertEqual(
            transformation.output_domain,
            SparkDataFrameDomain(
                {
                    "B": SparkStringColumnDescriptor(),
                    "A": SparkFloatColumnDescriptor(),
                    "C": SparkFloatColumnDescriptor(allow_nan=True, allow_inf=True),
                }
            ),
        )
        self.assertEqual(transformation.output_metric, SymmetricDifference())
        self.assertEqual(transformation.join_cols, ["B"])
        pd.testing.assert_frame_equal(
            transformation.public_df.toPandas(), self.public_df.toPandas()
        )
        self.assertEqual(transformation.stability, 2)

    @parameterized.expand(
        [
            (SymmetricDifference(), 2),
            (IfGroupedBy("B", SumOf(SymmetricDifference())), 2),
            (IfGroupedBy("B", RootSumOfSquared(SymmetricDifference())), 2),
            (IfGroupedBy("B", SymmetricDifference()), 1),
        ]
    )
    def test_public_join_correctness(
        self, metric: Union[SymmetricDifference, IfGroupedBy], d_out: int
    ):
        """Tests that public join works correctly."""
        public_join_transformation = PublicJoin(
            input_domain=self.input_domain,
            public_df=self.public_df,
            metric=metric,
            join_cols=["B"],
        )
        self.assertTrue(
            public_join_transformation.output_metric
            == metric
            == public_join_transformation.input_metric
        )
        self.assertEqual(public_join_transformation.stability_function(1), d_out)
        self.assertTrue(public_join_transformation.stability_relation(1, d_out))
        joined_df = public_join_transformation(self.private_df)
        self.assertEqual(
            joined_df.schema,
            cast(
                SparkDataFrameDomain, public_join_transformation.output_domain
            ).spark_schema,
        )
        actual = joined_df.toPandas()
        expected = pd.DataFrame(
            [[1.2, "X", 10.0], [1.2, "X", 11.0]], columns=["A", "B", "C"]
        )
        self.assert_frame_equal_with_sort(actual, expected)

    def test_public_join_overlapping_columns(self):
        """Tests that public join works when columns not used in join overlap."""
        public_df = self.spark.createDataFrame(
            pd.DataFrame(
                [["ABC", "X", 10.0], ["DEF", "X", 11.0]], columns=["A", "B", "C"]
            )
        )
        public_join_transformation = PublicJoin(
            input_domain=self.input_domain,
            metric=SymmetricDifference(),
            public_df=public_df,
            join_cols=["B"],
        )
        expected_df = pd.DataFrame(
            [[1.2, "ABC", "X", 10.0], [1.2, "DEF", "X", 11.0]],
            columns=["A_left", "A_right", "B", "C"],
        )
        actual_df = public_join_transformation(self.private_df).toPandas()
        self.assert_frame_equal_with_sort(actual_df, expected_df)

    @parameterized.expand(
        [
            (
                ["B", "C"],
                ["B"],
                "C",
                "C is an overlapping column but not a join key",
                SymmetricDifference(),
            ),
            (
                ["A", "B"],
                ["B"],
                "D",
                "Input metric .* and input domain .* are not compatible",
                SymmetricDifference(),
            ),
            (["A", "B"], ["B"], "A", "must be SymmetricDifference", HammingDistance()),
        ]
    )
    def test_if_grouped_by_metric_invalid_parameters(
        self,
        private_cols: List[str],
        join_cols: List[str],
        groupby_col: str,
        error_msg: str,
        inner_metric: Union[SymmetricDifference, HammingDistance],
    ):
        """Tests that PublicJoin raises appropriate errors with invalid params."""
        with self.assertRaisesRegex(ValueError, error_msg):
            PublicJoin(
                input_domain=SparkDataFrameDomain(
                    {col: SparkStringColumnDescriptor() for col in private_cols}
                ),
                public_df=self.spark.createDataFrame(
                    pd.DataFrame(
                        {"X": ["a1", "a2"], "C": ["z1", "z2"], "B": ["1", "2"]}
                    )
                ),
                metric=IfGroupedBy(groupby_col, SumOf(inner_metric)),
                join_cols=join_cols,
            )

    def test_join_with_mismatching_public_df_and_domain(self):
        """Tests that error is raised if public_df spark schema and domain mismatch."""
        with self.assertRaisesRegex(
            ValueError, "public_df's Spark schema does not match public_df_domain"
        ):
            PublicJoin(
                input_domain=self.input_domain,
                metric=SymmetricDifference(),
                public_df=self.spark.createDataFrame(
                    [("X", 10.0), ("X", 11.0)],
                    schema=st.StructType(
                        [
                            st.StructField("B", st.StringType(), nullable=True),
                            st.StructField("C", st.DoubleType(), nullable=True),
                        ]
                    ),
                ),
                public_df_domain=SparkDataFrameDomain(
                    {
                        "B": SparkStringColumnDescriptor(),
                        "C": SparkFloatColumnDescriptor(),
                    }
                ),
            )

    def test_join_with_public_df_domain(self):
        """Tests that join output domain is correctly inferred from public DF domain."""
        public_join = PublicJoin(
            input_domain=self.input_domain,
            metric=SymmetricDifference(),
            public_df=self.public_df,
            public_df_domain=SparkDataFrameDomain(
                {"B": SparkStringColumnDescriptor(), "C": SparkFloatColumnDescriptor()}
            ),
        )
        actual = public_join.output_domain
        expected = SparkDataFrameDomain(
            {
                "B": SparkStringColumnDescriptor(),
                "A": SparkFloatColumnDescriptor(),
                "C": SparkFloatColumnDescriptor(),
            }
        )
        self.assertEqual(actual, expected)

    def test_join_drops_invalid_rows_from_public_df(self):
        """ "Tests that nans/infs are dropped from public DataFrame when disallowed."""
        public_join = PublicJoin(
            input_domain=self.input_domain,
            metric=SymmetricDifference(),
            public_df=self.spark.createDataFrame(
                [("X", float("nan")), ("X", float("inf")), ("X", 1.1)],
                schema=st.StructType(
                    [
                        st.StructField("B", st.StringType(), nullable=False),
                        st.StructField("C", st.DoubleType(), nullable=False),
                    ]
                ),
            ),
            public_df_domain=SparkDataFrameDomain(
                {"B": SparkStringColumnDescriptor(), "C": SparkFloatColumnDescriptor()}
            ),
        )
        actual = public_join.public_df.toPandas()
        expected = pd.DataFrame({"B": ["X"], "C": [1.1]})
        self.assert_frame_equal_with_sort(actual, expected)

    @parameterized.expand(
        [
            (
                True,
                pd.DataFrame(
                    [["X", 1.2, 1.1], [None, 0.1, 1.2], [None, 0.1, 2.1]],
                    columns=["B", "A", "C"],
                ),
            ),
            (False, pd.DataFrame([["X", 1.2, 1.1]], columns=["B", "A", "C"])),
        ]
    )
    def test_join_null_behavior(self, join_on_nulls: bool, expected: pd.DataFrame):
        """Tests that PublicJoin deals with null values on join columns correctly."""
        public_join = PublicJoin(
            input_domain=SparkDataFrameDomain(
                {
                    "A": SparkFloatColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(allow_null=True),
                }
            ),
            metric=SymmetricDifference(),
            public_df=self.spark.createDataFrame(
                [(None, 2.1), (None, 1.2), ("X", 1.1)], schema=["B", "C"]
            ),
            public_df_domain=SparkDataFrameDomain(
                {
                    "B": SparkStringColumnDescriptor(allow_null=True),
                    "C": SparkFloatColumnDescriptor(allow_null=True),
                }
            ),
            join_on_nulls=join_on_nulls,
        )
        private_df = self.spark.createDataFrame(
            [(1.2, "X"), (0.1, None)], schema=["A", "B"]
        )
        actual = public_join(private_df).toPandas()
        self.assert_frame_equal_with_sort(actual, expected)

    def test_join_on_nulls_stability(self):
        """Tests that PublicJoin computes stability correctly when joining on nulls."""
        public_join = PublicJoin(
            input_domain=self.input_domain,
            metric=SymmetricDifference(),
            public_df=self.spark.createDataFrame(
                [(None, 2.1), (None, 1.2), ("X", 1.1)],
                schema=st.StructType(
                    [
                        st.StructField("B", st.StringType()),
                        st.StructField("C", st.DoubleType(), nullable=False),
                    ]
                ),
            ),
            public_df_domain=SparkDataFrameDomain(
                {
                    "B": SparkStringColumnDescriptor(allow_null=True),
                    "C": SparkFloatColumnDescriptor(),
                }
            ),
            join_on_nulls=True,
        )
        self.assertTrue(public_join.stability == 2)

    def test_join_stability_ignores_nulls(self):
        """Tests that stability is correct when join_on_nulls is False."""
        public_join = PublicJoin(
            input_domain=self.input_domain,
            metric=SymmetricDifference(),
            public_df=self.spark.createDataFrame(
                [(None, 2.1), (None, 1.2), ("X", 1.1)],
                schema=st.StructType(
                    [
                        st.StructField("B", st.StringType()),
                        st.StructField("C", st.DoubleType(), nullable=False),
                    ]
                ),
            ),
            public_df_domain=SparkDataFrameDomain(
                {
                    "B": SparkStringColumnDescriptor(allow_null=True),
                    "C": SparkFloatColumnDescriptor(),
                }
            ),
            join_on_nulls=False,
        )
        self.assertTrue(public_join.stability == 1)

    def test_empty_public_dataframe(self):
        """Tests that PublicJoin works with empty public DataFrame."""
        public_join_transformation = PublicJoin(
            input_domain=self.input_domain,
            metric=SymmetricDifference(),
            public_df=self.spark.createDataFrame([], schema=self.public_df.schema),
            join_cols=["B"],
        )
        actual = public_join_transformation(self.private_df).toPandas()
        expected = pd.DataFrame({"B": [], "A": [], "C": []})
        self.assert_frame_equal_with_sort(actual, expected)


class TestPrivateJoin(PySparkTest):
    """Tests for class PrivateJoin.

    Tests :class:`~tmlt.core.transformations.spark_transformations.join.PrivateJoin`.
    """

    def setUp(self):
        """Setup."""
        self.left_domain = SparkDataFrameDomain(
            {"A": SparkIntegerColumnDescriptor(), "B": SparkStringColumnDescriptor()}
        )
        self.right_domain = SparkDataFrameDomain(
            {"B": SparkStringColumnDescriptor(), "C": SparkStringColumnDescriptor()}
        )

    def test_constructor_mutable_arguments(self):
        """Tests that mutable constructor arguments are copied."""
        join_cols = ["B"]
        transformation = PrivateJoin(
            input_domain=DictDomain(
                {"l": self.left_domain, ("r", "i", "g", "h", "t"): self.right_domain}
            ),
            left_key="l",
            right_key=("r", "i", "g", "h", "t"),
            left_truncation_strategy=TruncationStrategy.TRUNCATE,
            right_truncation_strategy=TruncationStrategy.TRUNCATE,
            left_truncation_threshold=1,
            right_truncation_threshold=1,
            join_cols=join_cols,
        )
        join_cols.append("C")
        self.assertListEqual(transformation.join_cols, ["B"])

    @parameterized.expand(get_all_props(PrivateJoin))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        transformation = PrivateJoin(
            input_domain=DictDomain(
                {"l": self.left_domain, ("r", "i", "g", "h", "t"): self.right_domain}
            ),
            left_key="l",
            right_key=("r", "i", "g", "h", "t"),
            left_truncation_strategy=TruncationStrategy.TRUNCATE,
            right_truncation_strategy=TruncationStrategy.TRUNCATE,
            left_truncation_threshold=1,
            right_truncation_threshold=1,
            join_cols=["B"],
        )
        assert_property_immutability(transformation, prop_name)

    @parameterized.expand([(["B"], True), (None, False)])
    def test_properties(self, join_cols: Optional[List[str]], join_on_nulls: bool):
        """Tests that PrivateJoin's properties have expected values."""
        input_domain = DictDomain(
            {"l": self.left_domain, ("r", "i", "g", "h", "t"): self.right_domain}
        )
        transformation = PrivateJoin(
            input_domain=input_domain,
            left_key="l",
            right_key=("r", "i", "g", "h", "t"),
            left_truncation_strategy=TruncationStrategy.TRUNCATE,
            right_truncation_strategy=TruncationStrategy.TRUNCATE,
            left_truncation_threshold=1,
            right_truncation_threshold=2,
            join_cols=join_cols,
            join_on_nulls=join_on_nulls,
        )

        expected_output_metric = DictMetric(
            {
                "l": SymmetricDifference(),
                ("r", "i", "g", "h", "t"): SymmetricDifference(),
            }
        )
        expected_output_domain = SparkDataFrameDomain(
            {
                "B": SparkStringColumnDescriptor(),
                "A": SparkIntegerColumnDescriptor(),
                "C": SparkStringColumnDescriptor(),
            }
        )

        self.assertEqual(transformation.input_domain, input_domain)
        self.assertEqual(transformation.input_metric, expected_output_metric)
        self.assertEqual(transformation.output_domain, expected_output_domain)
        self.assertEqual(transformation.output_metric, SymmetricDifference())
        self.assertEqual(transformation.left_key, "l")
        self.assertEqual(transformation.right_key, ("r", "i", "g", "h", "t"))
        self.assertEqual(
            transformation.left_truncation_strategy, TruncationStrategy.TRUNCATE
        )
        self.assertEqual(
            transformation.right_truncation_strategy, TruncationStrategy.TRUNCATE
        )
        self.assertEqual(transformation.left_truncation_threshold, 1)
        self.assertEqual(transformation.right_truncation_threshold, 2)
        self.assertEqual(transformation.join_cols, ["B"])
        self.assertEqual(transformation.join_on_nulls, join_on_nulls)

    @parameterized.expand(
        [
            (left_cols, right_cols, join_cols, expected_ordering, join_on_nulls)
            for (left_cols, right_cols, join_cols, expected_ordering) in [
                (["A", "B", "C"], ["B", "D"], ["B"], ["B", "A", "C", "D"]),
                (
                    ["A", "B", "C"],
                    ["B", "D", "C"],
                    ["B"],
                    ["B", "A", "C_left", "D", "C_right"],
                ),
                (
                    ["A", "B", "C"],
                    ["B", "D", "C"],
                    ["B"],
                    ["B", "A", "C_left", "D", "C_right"],
                ),
                (["A", "B", "C"], ["B", "C", "D"], ["C", "B"], ["C", "B", "A", "D"]),
                (["A", "B"], ["B", "C"], ["B"], ["B", "A", "C"]),
            ]
            for join_on_nulls in [True, False]
        ]
    )
    def test_columns_ordering(
        self,
        left_cols: List[str],
        right_cols: List[str],
        join_cols: List[str],
        expected_ordering: List[str],
        join_on_nulls: bool,
    ):
        """Tests that the output columns of join are in expected order.

        This checks:
            - Join columns (in the order given by the user) appear first.
            - Columns of left table (with _left appended as required) appear
             next in the input order. (excluding join columns)
            - Columns of the right table (with _right appended as required) appear
             last in the input order. (excluding join columns)
        """
        left_domain = SparkDataFrameDomain(
            {col: SparkStringColumnDescriptor() for col in left_cols}
        )
        right_domain = SparkDataFrameDomain(
            {col: SparkStringColumnDescriptor() for col in right_cols}
        )

        left_df = self.spark.createDataFrame(
            [("x",) * len(left_cols)], schema=left_cols
        )
        right_df = self.spark.createDataFrame(
            [("x",) * len(right_cols)], schema=right_cols
        )

        private_join = PrivateJoin(
            input_domain=DictDomain({"left": left_domain, "right": right_domain}),
            left_key="left",
            right_key="right",
            left_truncation_strategy=TruncationStrategy.TRUNCATE,
            right_truncation_strategy=TruncationStrategy.TRUNCATE,
            left_truncation_threshold=1,
            right_truncation_threshold=1,
            join_cols=join_cols,
            join_on_nulls=join_on_nulls,
        )

        answer = private_join({"left": left_df, "right": right_df})
        self.assertTrue(answer in private_join.output_domain)
        self.assertEqual(answer.columns, expected_ordering)

    @parameterized.expand(
        [
            (1, 10, 22, TruncationStrategy.TRUNCATE),
            (5, 5, 20, TruncationStrategy.TRUNCATE),
            (1, 10, 20, TruncationStrategy.DROP),
            (5, 5, 50, TruncationStrategy.DROP),
        ]
    )
    def test_stability_relation(
        self,
        threshold_left: int,
        threshold_right: int,
        d_out: int,
        truncation_strategy: TruncationStrategy,
    ):
        """Tests that PrivateJoin's stability relation is correct."""
        join_transformation = PrivateJoin(
            input_domain=DictDomain(
                {"left": self.left_domain, "right": self.right_domain}
            ),
            left_key="left",
            right_key="right",
            left_truncation_strategy=truncation_strategy,
            right_truncation_strategy=truncation_strategy,
            left_truncation_threshold=threshold_left,
            right_truncation_threshold=threshold_right,
            join_cols=["B"],
        )
        self.assertEqual(
            join_transformation.stability_function({"left": 1, "right": 1}), d_out
        )

    @parameterized.expand(
        [
            (
                pd.DataFrame([(1, 2), (1, 3), (2, 4)], columns=["A", "B"]),
                pd.DataFrame([(2, 5), (1, 6)], columns=["A", "B"]),
                TruncationStrategy.TRUNCATE,
                2,
                ["A"],
                pd.DataFrame(
                    [(1, 2, 6), (1, 3, 6), (2, 4, 5)],
                    columns=["A", "B_left", "B_right"],
                ),
            ),
            (
                pd.DataFrame([(1, 2), (1, 3), (2, 4)], columns=["A", "B"]),
                pd.DataFrame([(2, 5), (1, 6)], columns=["A", "B"]),
                TruncationStrategy.DROP,
                1,
                ["A"],
                pd.DataFrame([(2, 4, 5)], columns=["A", "B_left", "B_right"]),
            ),
            (
                pd.DataFrame([(1, 2), (1, 3), (2, 4)], columns=["A", "B"]),
                pd.DataFrame([(2, 5), (2, 2), (1, 6)], columns=["A", "B"]),
                TruncationStrategy.DROP,
                1,
                ["A"],
                pd.DataFrame([], columns=["A", "B_left", "B_right"]),
            ),
        ]
    )
    def test_correctness(
        self,
        left: pd.DataFrame,
        right: pd.DataFrame,
        truncation_strategy: TruncationStrategy,
        threshold: int,
        join_cols: List[str],
        expected: pd.DataFrame,
    ):
        """Tests that join is computed correctly."""
        left_domain = SparkDataFrameDomain(
            {col: SparkIntegerColumnDescriptor() for col in left.columns}
        )
        right_domain = SparkDataFrameDomain(
            {col: SparkIntegerColumnDescriptor() for col in right.columns}
        )
        private_join = PrivateJoin(
            input_domain=DictDomain({"left": left_domain, "right": right_domain}),
            left_key="left",
            right_key="right",
            left_truncation_strategy=truncation_strategy,
            right_truncation_strategy=truncation_strategy,
            left_truncation_threshold=threshold,
            right_truncation_threshold=threshold,
            join_cols=join_cols,
        )
        left_sdf = self.spark.createDataFrame(left)
        right_sdf = self.spark.createDataFrame(right)
        actual = private_join({"left": left_sdf, "right": right_sdf}).toPandas()
        self.assert_frame_equal_with_sort(actual, expected)

    @parameterized.expand(
        [
            (  # Domain contains > 2 keys
                DictDomain(
                    {
                        "df1": SparkDataFrameDomain(
                            {"A": SparkIntegerColumnDescriptor()}
                        ),
                        "df2": SparkDataFrameDomain(
                            {"A": SparkIntegerColumnDescriptor()}
                        ),
                        "df3": SparkDataFrameDomain(
                            {"A": SparkIntegerColumnDescriptor()}
                        ),
                    }
                ),
                "df1",
                "df2",
                ["A"],
                "must be a DictDomain with 2 keys",
            ),
            (  # Invalid key
                DictDomain(
                    {
                        "df1": SparkDataFrameDomain(
                            {"A": SparkIntegerColumnDescriptor()}
                        ),
                        "df2": SparkDataFrameDomain(
                            {"A": SparkIntegerColumnDescriptor()}
                        ),
                    }
                ),
                "df3",
                "df1",
                ["A"],
                "Key 'df3' not in input domain",
            ),
            (  # Identical left and right
                DictDomain(
                    {
                        "df1": SparkDataFrameDomain(
                            {"A": SparkIntegerColumnDescriptor()}
                        ),
                        "df2": SparkDataFrameDomain(
                            {"A": SparkIntegerColumnDescriptor()}
                        ),
                    }
                ),
                "df1",
                "df1",
                ["A"],
                "Left and right keys must be distinct",
            ),
            (  # No common columns
                DictDomain(
                    {
                        "df1": SparkDataFrameDomain(
                            {"B": SparkStringColumnDescriptor()}
                        ),
                        "df2": SparkDataFrameDomain(
                            {"A": SparkIntegerColumnDescriptor()}
                        ),
                    }
                ),
                "df1",
                "df2",
                None,
                "No common columns",
            ),
            (  # Mismatching column types
                DictDomain(
                    {
                        "df1": SparkDataFrameDomain(
                            {"A": SparkStringColumnDescriptor()}
                        ),
                        "df2": SparkDataFrameDomain(
                            {"A": SparkIntegerColumnDescriptor()}
                        ),
                    }
                ),
                "df1",
                "df2",
                ["A"],
                "mismatching types on join column A",
            ),
            (  # _right column already exists
                DictDomain(
                    {
                        "df1": SparkDataFrameDomain(
                            {
                                "A": SparkStringColumnDescriptor(),
                                "B": SparkStringColumnDescriptor(),
                                "B_right": SparkStringColumnDescriptor(),
                            }
                        ),
                        "df2": SparkDataFrameDomain(
                            {
                                "A": SparkStringColumnDescriptor(),
                                "B": SparkStringColumnDescriptor(),
                            }
                        ),
                    }
                ),
                "df1",
                "df2",
                ["A"],
                "Join would rename overlapping column 'B' to an existing column name",
            ),
        ]
    )
    def test_invalid_arguments_rejected(
        self,
        input_domain: DictDomain,
        left: str,
        right: str,
        join_cols: Optional[List[str]],
        error_msg: str,
    ):
        """Tests that PrivateJoin cannot be constructed with invalid arguments."""
        with self.assertRaisesRegex(ValueError, error_msg):
            PrivateJoin(
                input_domain=input_domain,
                left_key=left,
                right_key=right,
                left_truncation_strategy=TruncationStrategy.TRUNCATE,
                right_truncation_strategy=TruncationStrategy.TRUNCATE,
                left_truncation_threshold=1,
                right_truncation_threshold=1,
                join_cols=join_cols,
            )

    def test_join_drops_invalid_rows(self):
        """Tests nans are dropped from the right-hand DataFrame when disallowed.

        Infinity should still be allowed.
        """
        left = pd.DataFrame([(1, 2), (1, 3), (2, 4)], columns=["A", "B"])
        right_sdf = self.spark.createDataFrame(
            [(1, 2, float("nan")), (1, 3, float("inf")), (2, 4, float(0))],
            schema=st.StructType(
                [
                    st.StructField("A", st.IntegerType(), nullable=False),
                    st.StructField("B", st.IntegerType(), nullable=False),
                    st.StructField("C", st.DoubleType(), nullable=False),
                ]
            ),
        )
        left_domain = SparkDataFrameDomain(
            {col: SparkIntegerColumnDescriptor() for col in left.columns}
        )
        right_domain = SparkDataFrameDomain(
            {
                "A": SparkIntegerColumnDescriptor(),
                "B": SparkIntegerColumnDescriptor(),
                "C": SparkFloatColumnDescriptor(allow_inf=True),
            }
        )
        join_cols = ["A", "B"]
        private_join = PrivateJoin(
            input_domain=DictDomain({"left": left_domain, "right": right_domain}),
            left_key="left",
            right_key="right",
            left_truncation_strategy=TruncationStrategy.TRUNCATE,
            right_truncation_strategy=TruncationStrategy.TRUNCATE,
            left_truncation_threshold=1,
            right_truncation_threshold=1,
            join_cols=join_cols,
            join_on_nulls=False,
        )
        left_sdf = self.spark.createDataFrame(left)
        actual = private_join({"left": left_sdf, "right": right_sdf}).toPandas()
        expected = pd.DataFrame(
            [[2, 4, float(0)], [1, 3, float("inf")]], columns=["A", "B", "C"]
        )
        self.assert_frame_equal_with_sort(actual, expected)

    def test_join_without_nulls_changes_domain(self):
        """Test that when join_on_null=False, output domain does not allow null."""
        left_domain = SparkDataFrameDomain(
            {
                "A": SparkFloatColumnDescriptor(),
                "B": SparkStringColumnDescriptor(allow_null=True),
            }
        )
        right_domain = SparkDataFrameDomain(
            {
                "B": SparkStringColumnDescriptor(allow_null=True),
                "C": SparkFloatColumnDescriptor(allow_null=True),
            }
        )
        private_join = PrivateJoin(
            input_domain=DictDomain({"left": left_domain, "right": right_domain}),
            left_key="left",
            right_key="right",
            left_truncation_strategy=TruncationStrategy.TRUNCATE,
            right_truncation_strategy=TruncationStrategy.TRUNCATE,
            left_truncation_threshold=10,
            right_truncation_threshold=10,
            join_on_nulls=False,
        )
        expected_output_domain = SparkDataFrameDomain(
            {
                "A": SparkFloatColumnDescriptor(),
                "B": SparkStringColumnDescriptor(allow_null=False),
                "C": SparkFloatColumnDescriptor(allow_null=True),
            }
        )
        actual = private_join.output_domain
        self.assertEqual(expected_output_domain["A"], actual["A"])
        self.assertEqual(expected_output_domain["B"], actual["B"])
        self.assertEqual(expected_output_domain["C"], actual["C"])

    @parameterized.expand(
        [
            (
                True,
                pd.DataFrame(
                    [["X", 1.2, 1.1], [None, 0.1, 1.2], [None, 0.1, 2.1]],
                    columns=["B", "A", "C"],
                ),
            ),
            (False, pd.DataFrame([["X", 1.2, 1.1]], columns=["B", "A", "C"])),
        ]
    )
    def test_join_on_nulls_behavior(self, join_on_nulls: bool, expected: pd.DataFrame):
        """Test that PrivateJoin deals with null values on join columns correctly."""
        left = self.spark.createDataFrame([(1.2, "X"), (0.1, None)], schema=["A", "B"])
        right = self.spark.createDataFrame(
            [(None, 2.1), (None, 1.2), ("X", 1.1)], schema=["B", "C"]
        )
        left_domain = SparkDataFrameDomain(
            {
                "A": SparkFloatColumnDescriptor(),
                "B": SparkStringColumnDescriptor(allow_null=True),
            }
        )
        right_domain = SparkDataFrameDomain(
            {
                "B": SparkStringColumnDescriptor(allow_null=True),
                "C": SparkFloatColumnDescriptor(allow_null=True),
            }
        )
        private_join = PrivateJoin(
            input_domain=DictDomain({"left": left_domain, "right": right_domain}),
            left_key="left",
            right_key="right",
            left_truncation_strategy=TruncationStrategy.DROP,
            right_truncation_strategy=TruncationStrategy.DROP,
            left_truncation_threshold=10,
            right_truncation_threshold=10,
            join_on_nulls=join_on_nulls,
        )
        actual = private_join({"left": left, "right": right}).toPandas()
        self.assert_frame_equal_with_sort(actual, expected)


class TestPrivateJoinOnKey(PySparkTest):
    """Tests for class PrivateJoinOnKey.

    Tests :class:`~tmlt.core.transformations.spark_transformations.join.PrivateJoinOnKey`.
    """  # pylint: disable=line-too-long

    def setUp(self):
        """Setup."""
        self.left_domain = SparkDataFrameDomain(
            {"A": SparkIntegerColumnDescriptor(), "B": SparkStringColumnDescriptor()}
        )
        self.right_domain = SparkDataFrameDomain(
            {"B": SparkStringColumnDescriptor(), "C": SparkStringColumnDescriptor()}
        )

    def test_constructor_mutable_arguments(self):
        """Tests that mutable constructor arguments are copied."""
        join_cols = ["B"]
        transformation = PrivateJoinOnKey(
            input_domain=DictDomain(
                {"l": self.left_domain, ("r", "i", "g", "h", "t"): self.right_domain}
            ),
            input_metric=AddRemoveKeys({"l": "B", ("r", "i", "g", "h", "t"): "B"}),
            left_key="l",
            right_key=("r", "i", "g", "h", "t"),
            new_key="joined",
            join_cols=join_cols,
        )
        join_cols.append("C")
        self.assertListEqual(transformation.join_cols, ["B"])

    @parameterized.expand(get_all_props(PrivateJoinOnKey))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        transformation = PrivateJoinOnKey(
            input_domain=DictDomain(
                {"l": self.left_domain, ("r", "i", "g", "h", "t"): self.right_domain}
            ),
            input_metric=AddRemoveKeys({"l": "B", ("r", "i", "g", "h", "t"): "B"}),
            left_key="l",
            right_key=("r", "i", "g", "h", "t"),
            new_key="joined",
            join_cols=["B"],
        )
        assert_property_immutability(transformation, prop_name)

    @parameterized.expand([(["B"], True), (None, False)])
    def test_properties(self, join_cols: Optional[List[str]], join_on_nulls: bool):
        """Tests that PrivateJoinOnKey's properties have expected values."""
        input_domain = DictDomain(
            {"l": self.left_domain, ("r", "i", "g", "h", "t"): self.right_domain}
        )
        transformation = PrivateJoinOnKey(
            input_domain=DictDomain(
                {"l": self.left_domain, ("r", "i", "g", "h", "t"): self.right_domain}
            ),
            input_metric=AddRemoveKeys({"l": "B", ("r", "i", "g", "h", "t"): "B"}),
            left_key="l",
            right_key=("r", "i", "g", "h", "t"),
            new_key="joined",
            join_cols=join_cols,
            join_on_nulls=join_on_nulls,
        )

        expected_output_metric = AddRemoveKeys(
            {"l": "B", ("r", "i", "g", "h", "t"): "B"}
        )
        expected_output_domain = DictDomain(
            {
                "l": self.left_domain,
                ("r", "i", "g", "h", "t"): self.right_domain,
                "joined": SparkDataFrameDomain(
                    {
                        "B": SparkStringColumnDescriptor(),
                        "A": SparkIntegerColumnDescriptor(),
                        "C": SparkStringColumnDescriptor(),
                    }
                ),
            }
        )

        self.assertEqual(transformation.input_domain, input_domain)
        self.assertEqual(transformation.input_metric, expected_output_metric)
        self.assertEqual(transformation.output_domain, expected_output_domain)
        self.assertEqual(
            transformation.output_metric,
            AddRemoveKeys({"l": "B", ("r", "i", "g", "h", "t"): "B", "joined": "B"}),
        )
        self.assertEqual(transformation.left_key, "l")
        self.assertEqual(transformation.right_key, ("r", "i", "g", "h", "t"))
        self.assertEqual(transformation.new_key, "joined")
        self.assertEqual(transformation.join_cols, ["B"])
        self.assertEqual(transformation.join_on_nulls, join_on_nulls)

    @parameterized.expand(
        [
            (left_cols, right_cols, join_cols, expected_ordering, join_on_nulls)
            for (left_cols, right_cols, join_cols, expected_ordering) in [
                (["A", "B", "C"], ["B", "D"], ["B"], ["B", "A", "C", "D"]),
                (
                    ["A", "B", "C"],
                    ["B", "D", "C"],
                    ["B"],
                    ["B", "A", "C_left", "D", "C_right"],
                ),
                (
                    ["A", "B", "C"],
                    ["B", "D", "C"],
                    ["B"],
                    ["B", "A", "C_left", "D", "C_right"],
                ),
                (["A", "B", "C"], ["B", "C", "D"], ["C", "B"], ["C", "B", "A", "D"]),
                (["A", "B"], ["B", "C"], ["B"], ["B", "A", "C"]),
            ]
            for join_on_nulls in [True, False]
        ]
    )
    def test_columns_ordering(
        self,
        left_cols: List[str],
        right_cols: List[str],
        join_cols: List[str],
        expected_ordering: List[str],
        join_on_nulls: bool,
    ):
        """Tests that the output columns of join are in expected order.

        This checks:
            - Join columns (in the order given by the user) appear first.
            - Columns of left table (with _left appended as required) appear
             next in the input order. (excluding join columns)
            - Columns of the right table (with _right appended as required) appear
             last in the input order. (excluding join columns)
        """
        left_domain = SparkDataFrameDomain(
            {col: SparkStringColumnDescriptor() for col in left_cols}
        )
        right_domain = SparkDataFrameDomain(
            {col: SparkStringColumnDescriptor() for col in right_cols}
        )

        left_df = self.spark.createDataFrame(
            [("x",) * len(left_cols)], schema=left_cols
        )
        right_df = self.spark.createDataFrame(
            [("x",) * len(right_cols)], schema=right_cols
        )

        private_join = PrivateJoinOnKey(
            input_domain=DictDomain({"left": left_domain, "right": right_domain}),
            input_metric=AddRemoveKeys({"left": "B", "right": "B"}),
            left_key="left",
            right_key="right",
            new_key="joined",
            join_cols=join_cols,
            join_on_nulls=join_on_nulls,
        )

        answer = private_join({"left": left_df, "right": right_df})
        self.assertTrue(answer in private_join.output_domain)
        self.assertEqual(answer["joined"].columns, expected_ordering)

    @parameterized.expand(
        [(1, 1, True), (0, 0, True), (float("inf"), float("inf"), True), (2, 1, False)]
    )
    def test_stability_relation_and_function(
        self, d_in: ExactNumberInput, d_out: ExactNumberInput, expected: bool
    ):
        """Test that PrivateJoinOnKey's stability relation and function are correct"""
        private_join = PrivateJoinOnKey(
            input_domain=DictDomain(
                {"left": self.left_domain, "right": self.right_domain}
            ),
            input_metric=AddRemoveKeys({"left": "B", "right": "B"}),
            left_key="left",
            right_key="right",
            new_key="new",
            join_cols=["B"],
        )
        self.assertEqual(private_join.stability_relation(d_in, d_out), expected)
        self.assertEqual(private_join.stability_function(d_in) <= d_out, expected)



"""Unit tests for :mod:`~tmlt.core.domains.collections`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2022
from typing import Any, Dict, Optional
from unittest.case import TestCase
from unittest.mock import Mock, create_autospec

import numpy as np
from parameterized import parameterized
from pyspark.sql.types import StringType

from tmlt.core.domains.base import Domain, OutOfDomainError
from tmlt.core.domains.collections import DictDomain, ListDomain
from tmlt.core.domains.numpy_domains import NumpyFloatDomain, NumpyIntegerDomain
from tmlt.core.utils.testing import assert_property_immutability, get_all_props


class TestListDomain(TestCase):
    """Tests for :class:`~tmlt.core.domains.collections.ListDomain`."""

    def setUp(self):
        """Setup."""
        self.list_domain = ListDomain(NumpyIntegerDomain())

    @parameterized.expand(
        [
            (NumpyFloatDomain(), False),
            (ListDomain(NumpyFloatDomain()), False),
            (ListDomain(NumpyIntegerDomain()), True),
        ]
    )
    def test_eq(self, domain: Domain, equal_domain: bool):
        """Tests that __eq__  works correctly."""
        self.assertEqual(self.list_domain == domain, equal_domain)

    @parameterized.expand([(None,), (StringType,)])
    def test_invalid_inputs(self, element_domain: Domain):
        """Test ListDomain with invalid input."""
        with self.assertRaises(TypeError):
            ListDomain(element_domain)

    @parameterized.expand(
        [
            ([np.int64(1)], None),
            ("Not a list", f"Value must be {list}, instead it is {str}."),
            (
                ["invalid"],
                f"Found invalid value in list: Value must be {np.int64}, "
                f"instead it is {str}.",
            ),
        ]
    )
    def test_validate(self, candidate: Any, exception: Optional[str]):
        """Tests that validate works correctly."""
        if exception is not None:
            with self.assertRaisesRegex(OutOfDomainError, exception):
                self.list_domain.validate(candidate)
        else:
            self.assertEqual(self.list_domain.validate(candidate), exception)


class TestDictDomain(TestCase):
    """Tests for :class:`~tmlt.core.domains.collections.DictDomain`."""

    def setUp(self):
        """Setup."""
        self.domain_a = create_autospec(spec=Domain, instance=True)
        self.domain_b = create_autospec(spec=Domain, instance=True)
        self.dict_domain = DictDomain({"A": self.domain_a, "B": self.domain_b})

    def test_constructor_mutable_arguments(self):
        """Tests that mutable constructor arguments are copied."""
        domain_map = {"A": NumpyIntegerDomain()}
        domain = DictDomain(key_to_domain=domain_map)
        domain_map["A"] = NumpyFloatDomain()
        self.assertDictEqual(domain.key_to_domain, {"A": NumpyIntegerDomain()})

    @parameterized.expand(get_all_props(DictDomain))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        assert_property_immutability(self.dict_domain, prop_name)

    @parameterized.expand(
        [
            ({"A": Mock(), "B": Mock()}, True, True),
            ({"A": Mock(), "B": Mock()}, True, False),
            ({"A": Mock(), "B": Mock()}, False, True),
            ({"A": Mock(), "B": Mock()}, False, False),
            ({"C": Mock(), "B": Mock()}, True, True),
            ({"A": Mock(), "B": Mock(), "C": Mock()}, True, True),
            ({"A": Mock()}, True, True),
        ]
    )
    def test_validate(self, candidate: Dict[str, Any], in_A: bool, in_B: bool):
        """Tests that validate works correctly."""
        self.domain_a.validate = (
            Mock(side_effect=OutOfDomainError("Test"))
            if not in_A
            else Mock(return_value=None)
        )
        self.domain_b.validate = (
            Mock(side_effect=OutOfDomainError("Test"))
            if not in_B
            else Mock(return_value=None)
        )

        if (in_A and in_B) and set(candidate) == {"A", "B"}:
            self.dict_domain.validate(candidate)
        else:
            issue_object = "'B'" if in_A else "'A'"
            exception = f"Found invalid value at {issue_object}: Test"
            if set(candidate) != {"A", "B"}:
                exception = (
                    "Keys are not as expected, value must match domain.\n"
                    fr"Value keys: \[{str(sorted(set(candidate)))[1:-1]}\]"
                    "\n"
                    r"Domain keys: \['A', 'B'\]"
                )
            with self.assertRaisesRegex(OutOfDomainError, exception):
                self.dict_domain.validate(candidate)

        if set(candidate) == {"A", "B"}:
            self.domain_a.validate.assert_called_once_with(candidate["A"])
            if in_A:
                self.domain_b.validate.assert_called_once_with(candidate["B"])

    @parameterized.expand(
        [
            (DictDomain({"A": NumpyIntegerDomain(), "B": NumpyFloatDomain()}), True),
            (DictDomain({"B": NumpyFloatDomain(), "A": NumpyIntegerDomain()}), True),
            (
                DictDomain(
                    {"A": NumpyIntegerDomain(), "B": NumpyFloatDomain(allow_nan=True)}
                ),
                False,
            ),
            (DictDomain({"A": NumpyIntegerDomain(), "C": NumpyFloatDomain()}), False),
            (
                DictDomain(
                    {
                        "A": NumpyIntegerDomain(),
                        "B": NumpyFloatDomain(),
                        "C": NumpyFloatDomain(),
                    }
                ),
                False,
            ),
            (DictDomain({"A": NumpyIntegerDomain()}), False),
            (NumpyIntegerDomain(), False),
        ]
    )
    def test_eq(self, candidate: Domain, expected: bool):
        """Tests that __eq__ works correctly."""
        domain = DictDomain({"A": NumpyIntegerDomain(), "B": NumpyFloatDomain()})
        self.assertEqual(domain == candidate, expected)

    def test_repr(self):
        """Tests that __repr__ works correctly"""
        domain = DictDomain({"A": NumpyIntegerDomain(), "B": NumpyFloatDomain()})

        expected = (
            "DictDomain(key_to_domain={'A': NumpyIntegerDomain(size=64), "
            "'B': NumpyFloatDomain(allow_nan=False, allow_inf=False, size=64)})"
        )
        self.assertEqual(repr(domain), expected)



"""Unit tests for :mod:`~tmlt.core.measurements.noise_mechanisms`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2022

# pylint: disable=no-self-use

import numpy as np
import sympy as sp
from parameterized import parameterized

from tmlt.core.domains.numpy_domains import NumpyFloatDomain, NumpyIntegerDomain
from tmlt.core.measurements.noise_mechanisms import (
    AddDiscreteGaussianNoise,
    AddGeometricNoise,
    AddLaplaceNoise,
)
from tmlt.core.measures import PureDP, RhoZCDP
from tmlt.core.metrics import AbsoluteDifference
from tmlt.core.utils.testing import (
    TestComponent,
    assert_property_immutability,
    get_all_props,
)


class TestAddLaplaceNoise(TestComponent):
    """Tests for :class:`~tmlt.core.measurements.noise_mechanisms.AddLaplaceNoise`."""

    @parameterized.expand(get_all_props(AddLaplaceNoise))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        measurement = AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=1)
        assert_property_immutability(measurement, prop_name)

    def test_properties(self):
        """AddLaplaceNoise's properties have the expected values."""
        measurement = AddLaplaceNoise(
            input_domain=NumpyIntegerDomain(), scale=sp.Rational("0.3")
        )
        self.assertEqual(measurement.input_domain, NumpyIntegerDomain())
        self.assertEqual(measurement.input_metric, AbsoluteDifference())
        self.assertEqual(measurement.output_measure, PureDP())
        self.assertEqual(measurement.is_interactive, False)
        self.assertEqual(measurement.scale, sp.Rational("0.3"))

    @parameterized.expand([(-0.4,), (np.nan,), ("invalid",)])
    def test_sigma_validity(self, sigma):
        """Tests that invalid scale values are rejected."""
        with self.assertRaises((ValueError, TypeError)):
            AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=sigma)

    def test_no_noise(self):
        """Works correctly with no noise."""
        measurement = AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=0)
        self.assertEqual(measurement.privacy_function(1), sp.oo)
        self.assertTrue(measurement.privacy_relation(1, sp.oo))
        self.assertFalse(measurement.privacy_relation(1, sp.Pow(10, 6)))
        self.assertEqual(measurement(5.0), 5.0)

    def test_some_noise(self):
        """Works correctly with some noise."""
        measurement = AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=2)
        self.assertEqual(measurement.privacy_function(1), sp.Rational("0.5"))
        self.assertTrue(measurement.privacy_relation(1, sp.Rational("0.5")))
        self.assertFalse(measurement.privacy_relation(1, sp.Rational("0.49999999")))

    def test_infinite_noise(self):
        """Works correctly with infinite noise."""
        measurement = AddLaplaceNoise(input_domain=NumpyIntegerDomain(), scale=sp.oo)
        self.assertEqual(measurement.privacy_function(1), 0)
        self.assertTrue(measurement.privacy_relation(1, 0))
        self.assertTrue(measurement.privacy_relation(1, 1))
        # Equally likely to return -inf or inf
        self.assertTrue(np.isinf(measurement(5.0)))

    @parameterized.expand(
        [(NumpyFloatDomain(allow_inf=True),), (NumpyFloatDomain(allow_nan=True),)]
    )
    def test_input_domain(self, domain: NumpyFloatDomain):
        """Tests that input domain is checked correctly.

        Condition: Must be a *NumpyIntegerDomain* or a
         (non-nullable) *NumpyFloatDomain*
        """
        with self.assertRaisesRegex(
            ValueError, "Input domain must not contain infs or nans"
        ):
            AddLaplaceNoise(input_domain=domain, scale=1)

    @parameterized.expand(
        [
            (1, 0.5, 0),
            (2, 0.5, 0),
            (3.1415, 0.5, 0),
            (1, 0.75, 0.693147),
            (1, 0.25, -0.693147),
            (3.5, 0.86, 4.45538),
            (3.5, 0.14, -4.45538),
        ]
    )
    def test_inverse_cdf(self, scale, p, expected):
        """Tests that the inverse CDF is calculated correctly."""
        self.assertAlmostEqual(
            AddLaplaceNoise.inverse_cdf(scale=scale, probability=p), expected, places=6
        )


class TestAddGeometricNoise(TestComponent):
    """Tests for class AddGeometricNoise.

    Tests :class:`~tmlt.core.measurements.noise_mechanisms.AddGeometricNoise`.
    """

    @parameterized.expand(get_all_props(AddGeometricNoise))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        measurement = AddGeometricNoise(alpha=0)
        assert_property_immutability(measurement, prop_name)

    def test_properties(self):
        """AddGeometricNoise's properties have the expected values."""
        measurement = AddGeometricNoise(alpha=sp.Rational("0.5"))
        self.assertEqual(measurement.input_domain, NumpyIntegerDomain())
        self.assertEqual(measurement.input_metric, AbsoluteDifference())
        self.assertEqual(measurement.output_measure, PureDP())
        self.assertEqual(measurement.is_interactive, False)
        self.assertEqual(measurement.alpha, sp.Rational("0.5"))

    @parameterized.expand([(-0.4,), (np.nan,), ("invalid",)])
    def test_alpha_validity(self, alpha):
        """Tests that invalid alpha values are rejected."""
        with self.assertRaises((ValueError, TypeError)):
            AddGeometricNoise(alpha=alpha)

    def test_no_noise(self):
        """Works correctly with no noise."""
        measurement = AddGeometricNoise(alpha=0)
        self.assertTrue(measurement.privacy_function(1), sp.oo)
        self.assertTrue(measurement.privacy_relation(1, sp.oo))
        self.assertFalse(measurement.privacy_relation(1, sp.Pow(10, 6)))
        self.assertEqual(measurement(5), 5)

    def test_some_noise(self):
        """Works correctly with some noise."""
        measurement = AddGeometricNoise(alpha=2)
        self.assertTrue(measurement.privacy_function(1), sp.Rational("0.5"))
        self.assertTrue(measurement.privacy_relation(1, sp.Rational("0.5")))
        self.assertFalse(measurement.privacy_relation(1, sp.Rational("0.49")))

    def test_infinite_noise(self):
        """Raises an error with infinite noise."""
        with self.assertRaisesRegex(
            ValueError, "Invalid alpha: oo is not strictly less than inf"
        ):
            AddGeometricNoise(alpha=sp.oo)

    @parameterized.expand(
        [
            (1, 0.5, 0),
            (2, 0.5, 0),
            (3.1415, 0.5, 0),
            (1, 0.75, 1),
            (1, 0.25, -1),
            (3.5, 0.86, 4),
            (3.5, 0.14, -4),
        ]
    )
    def test_inverse_cdf(self, alpha, p, expected):
        """Tests that the inverse CDF is calculated correctly."""
        self.assertAlmostEqual(
            AddGeometricNoise.inverse_cdf(
                alpha=alpha, probability=p
            ),  # pylint: disable=protected-access
            expected,
            places=5,
        )


class TestAddDiscreteGaussianNoise(TestComponent):
    """Tests for class AddDiscreteGaussianNoise.

    Tests :class:`~tmlt.core.measurements.noise_mechanisms.AddDiscreteGaussianNoise`.
    """

    @parameterized.expand(get_all_props(AddDiscreteGaussianNoise))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        measurement = AddDiscreteGaussianNoise(sigma_squared=0)
        assert_property_immutability(measurement, prop_name)

    def test_properties(self):
        """AddDiscreteGaussianNoise's properties have the expected values."""
        measurement = AddDiscreteGaussianNoise(sigma_squared=sp.Rational("0.5"))
        self.assertEqual(measurement.input_domain, NumpyIntegerDomain())
        self.assertEqual(measurement.input_metric, AbsoluteDifference())
        self.assertEqual(measurement.output_measure, RhoZCDP())
        self.assertEqual(measurement.is_interactive, False)
        self.assertEqual(measurement.sigma_squared, sp.Rational("0.5"))

    @parameterized.expand([(-0.4,), (np.nan,), ("invalid",)])
    def test_sigma_squared_validity(self, sigma_squared):
        """Tests that invalid sigma_squared values are rejected."""
        with self.assertRaises((ValueError, TypeError)):
            AddDiscreteGaussianNoise(sigma_squared)

    def test_no_noise(self):
        """Works correctly with no noise."""
        measurement = AddDiscreteGaussianNoise(sigma_squared=0)
        self.assertTrue(measurement.privacy_function(1), sp.oo)
        self.assertTrue(measurement.privacy_relation(1, sp.oo))
        self.assertFalse(measurement.privacy_relation(1, sp.Pow(10, 6)))
        self.assertEqual(measurement(5), 5)

    def test_some_noise(self):
        """Works correctly with some noise."""
        measurement = AddDiscreteGaussianNoise(sigma_squared=2)
        self.assertTrue(measurement.privacy_function(1), sp.Rational("0.5"))
        self.assertTrue(measurement.privacy_relation(1, sp.Rational("0.25")))
        self.assertFalse(measurement.privacy_relation(1, sp.Rational("0.2499")))

    def test_infinite_noise(self):
        """Raises an error with infinite noise."""
        with self.assertRaisesRegex(
            ValueError, "Invalid sigma_squared: oo is not strictly less than inf"
        ):
            AddDiscreteGaussianNoise(sigma_squared=sp.oo)

    def test_detailed_fraction(self):
        """Works correctly with fractions that have high numerators/denominators.

        Test for bug #964.
        """
        for _ in range(10):  # Unfortunately, the failure was somewhat flaky.
            AddDiscreteGaussianNoise(sigma_squared=sp.Rational("0.9999999999999999"))(0)

    @parameterized.expand(
        [
            (1, 0.5, 0),
            (2, 0.5, 0),
            (3.1415, 0.5, 0),
            (1, 0.75, 1),
            (1, 0.25, -1),
            (3.5, 0.86, 2),
            (3.5, 0.14, -2),
        ]
    )
    def test_inverse_cdf(self, sigma_squared, p, expected):
        """Tests that the inverse CDF is calculated correctly."""
        self.assertAlmostEqual(
            AddDiscreteGaussianNoise.inverse_cdf(
                sigma_squared=sigma_squared, probability=p
            ),  # pylint: disable=protected-access
            expected,
            places=5,
        )



import subprocess
import sys
from pathlib import Path

build_dir = Path(__file__).parent
build_command = ["bash", str(build_dir / "build.sh")]
try:
    subprocess.run(build_command, check=True)
except subprocess.CalledProcessError:
    print("=" * 80)
    print("Failed to build C dependencies, see above output for details.")
    print("=" * 80)
    sys.exit(1)



"""Unit tests for :mod:`~tmlt.core.domains.spark_domains`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2022

import datetime
from typing import Any, Optional

import pandas as pd
from parameterized import parameterized
from pyspark.sql.types import (
    DateType,
    FloatType,
    IntegerType,
    LongType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)

from tmlt.core.domains.base import Domain, OutOfDomainError
from tmlt.core.domains.collections import ListDomain
from tmlt.core.domains.numpy_domains import (
    NumpyFloatDomain,
    NumpyIntegerDomain,
    NumpyStringDomain,
)
from tmlt.core.domains.spark_domains import (
    SparkColumnDescriptor,
    SparkColumnsDescriptor,
    SparkDataFrameDomain,
    SparkDateColumnDescriptor,
    SparkFloatColumnDescriptor,
    SparkGroupedDataFrameDomain,
    SparkIntegerColumnDescriptor,
    SparkRowDomain,
    SparkStringColumnDescriptor,
    SparkTimestampColumnDescriptor,
)
from tmlt.core.utils.grouped_dataframe import GroupedDataFrame
from tmlt.core.utils.testing import (
    PySparkTest,
    assert_property_immutability,
    get_all_props,
)


class TestSparkDataFrameDomain(PySparkTest):
    """Tests for :class:`SparkDataFrameDomain`."""

    def setUp(self):
        """Setup."""
        self.domain = SparkDataFrameDomain(
            schema={
                "A": SparkIntegerColumnDescriptor(),
                "B": SparkStringColumnDescriptor(),
            }
        )

    def test_constructor_mutable_arguments(self):
        """Tests that mutable constructor arguments are copied."""
        schema = {
            "A": SparkIntegerColumnDescriptor(),
            "B": SparkStringColumnDescriptor(),
        }
        domain = SparkDataFrameDomain(schema=schema)
        schema["A"] = NumpyFloatDomain()
        self.assertDictEqual(
            domain.schema,
            {"A": SparkIntegerColumnDescriptor(), "B": SparkStringColumnDescriptor()},
        )

    @parameterized.expand(get_all_props(SparkDataFrameDomain))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        assert_property_immutability(self.domain, prop_name)

    def test_repr(self):
        """Tests that __repr__ works correctly."""
        expected = (
            "SparkDataFrameDomain(schema={'A':"
            " SparkIntegerColumnDescriptor(allow_null=False, size=64), 'B':"
            " SparkStringColumnDescriptor(allow_null=False)})"
        )
        print(repr(self.domain))
        self.assertEqual(repr(self.domain), expected)


class TestSparkRowDomain(PySparkTest):
    """Tests for :class:`SparkRowDomain`."""

    def setUp(self):
        """Setup."""
        self.domain = SparkRowDomain(
            schema={
                "A": SparkIntegerColumnDescriptor(),
                "B": SparkStringColumnDescriptor(),
            }
        )

    def test_constructor_mutable_arguments(self):
        """Tests that mutable constructor arguments are copied."""
        schema = {
            "A": SparkIntegerColumnDescriptor(),
            "B": SparkStringColumnDescriptor(),
        }
        domain = SparkRowDomain(schema=schema)
        schema["A"] = NumpyFloatDomain()
        self.assertDictEqual(
            domain.schema,
            {"A": SparkIntegerColumnDescriptor(), "B": SparkStringColumnDescriptor()},
        )

    @parameterized.expand(get_all_props(SparkRowDomain))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""

        assert_property_immutability(self.domain, prop_name)

    def test_repr(self):
        """Tests that __repr__ works correctly."""
        expected = (
            "SparkRowDomain(schema={'A': SparkIntegerColumnDescriptor(allow_null=False,"
            " size=64), 'B': SparkStringColumnDescriptor(allow_null=False)})"
        )
        self.assertEqual(repr(self.domain), expected)


class TestSparkBasedDomains(PySparkTest):
    """Tests for Spark-based Domains.

    In particular, the following domains are tested:
        1. SparkDataFrameDomain
        2. SparkRowDomain
        3. ListDomain[SparkRowDomain]
    """

    def setUp(self):
        """Setup Schema."""

        self.schema = {
            "A": SparkStringColumnDescriptor(),
            "B": SparkStringColumnDescriptor(),
            "C": SparkFloatColumnDescriptor(),
        }

    @parameterized.expand(
        [
            (SparkDataFrameDomain, StringType),
            (SparkRowDomain, int),
            (SparkRowDomain, StringType),
        ]
    )
    def test_invalid_spark_domain_inputs(self, SparkDomain: type, invalid_input: type):
        """Test Spark-based domains with invalid inputs."""
        with self.assertRaises(TypeError):
            SparkDomain(invalid_input)

    @parameterized.expand(
        [
            (  # LongType() instead of DoubleType()
                SparkDataFrameDomain,
                pd.DataFrame(
                    [["A", "B", 10], ["V", "E", 12], ["A", "V", 13]],
                    columns=["A", "B", "C"],
                ),
                "Found invalid value in column 'C': Column must be"
                r" DoubleType(\(\))?, instead it is LongType(\(\))?.",
            ),
            (  # Missing Columns
                SparkDataFrameDomain,
                pd.DataFrame([["A", "B"], ["V", "E"], ["A", "V"]], columns=["A", "B"]),
                "Columns are not as expected. DataFrame and Domain must contain the "
                "same columns in the same order.\n"
                r"DataFrame columns: \['A', 'B'\]"
                "\n"
                r"Domain columns: \['A', 'B', 'C'\]",
            ),
            (
                SparkDataFrameDomain,
                pd.DataFrame(
                    [["A", "B", 1.1], ["V", "E", 1.2], ["A", "V", 1.3]],
                    columns=["A", "B", "C"],
                ),
                None,
            ),
        ]
    )
    def test_validate(
        self, SparkDomain: type, candidate: Any, exception: Optional[str]
    ):
        """Tests that validate works as expected.

        Args:
            SparkDomain: Domain type to be checked.
            candidate: Object to be checked for membership.
            exception: Expected exception if validation fails.
        """
        domain = (
            SparkDomain(self.schema)
            if SparkDomain != ListDomain
            else SparkDomain(SparkRowDomain(self.schema))
        )
        if isinstance(candidate, pd.DataFrame):
            candidate = self.spark.createDataFrame(candidate)

        if exception is not None:
            with self.assertRaisesRegex(OutOfDomainError, exception):
                domain.validate(candidate)
        else:
            self.assertEqual(domain.validate(candidate), exception)

    @parameterized.expand(
        [
            (  # matching
                {
                    "A": SparkStringColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                    "C": SparkFloatColumnDescriptor(),
                },
                True,
            ),
            (  # shuffled
                {
                    "B": SparkStringColumnDescriptor(),
                    "C": SparkFloatColumnDescriptor(),
                    "A": SparkStringColumnDescriptor(),
                },
                False,
            ),
            (  # Mismatching Types
                {
                    "A": SparkStringColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                    "C": SparkFloatColumnDescriptor(size=32),
                },
                False,
            ),
            (  # Extra attribute
                {
                    "A": SparkStringColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                    "C": SparkFloatColumnDescriptor(),
                    "D": SparkFloatColumnDescriptor(),
                },
                False,
            ),
            (  # Missing attribute
                {
                    "A": SparkStringColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                },
                False,
            ),
        ]
    )
    def test_eq(self, other_schema: SparkColumnsDescriptor, is_equal: bool):
        """Tests that __eq__ works as expected for SparkDataFrameDomain."""
        self.assertEqual(
            SparkDataFrameDomain(other_schema) == SparkDataFrameDomain(self.schema),
            is_equal,
        )
        self.assertEqual(
            SparkRowDomain(other_schema) == SparkRowDomain(self.schema), is_equal
        )
        self.assertEqual(
            ListDomain(SparkRowDomain(other_schema))
            == ListDomain(SparkRowDomain(self.schema)),
            is_equal,
        )


class TestSparkGroupedDataFrameDomain(PySparkTest):
    """Tests for SparkGroupedDataFrameDomain."""

    def setUp(self):
        """Setup test."""
        self.group_keys = self.spark.createDataFrame(
            [(1, "W"), (2, "X"), (3, "Y")], schema=["A", "B"]
        )
        self.schema = {
            "A": SparkIntegerColumnDescriptor(allow_null=True),
            "B": SparkStringColumnDescriptor(allow_null=True),
            "C": SparkIntegerColumnDescriptor(allow_null=True),
        }
        self.domain = SparkGroupedDataFrameDomain(
            schema=self.schema, group_keys=self.group_keys
        )

    def test_constructor_mutable_arguments(self):
        """Tests that mutable constructor arguments are copied."""
        schema = {
            "A": SparkIntegerColumnDescriptor(),
            "B": SparkStringColumnDescriptor(),
        }
        domain = SparkGroupedDataFrameDomain(schema=schema, group_keys=self.group_keys)
        schema["A"] = NumpyFloatDomain()
        self.assertDictEqual(
            domain.schema,
            {"A": SparkIntegerColumnDescriptor(), "B": SparkStringColumnDescriptor()},
        )

    @parameterized.expand(get_all_props(SparkGroupedDataFrameDomain))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        assert_property_immutability(self.domain, prop_name)

    def test_carrier_type(self):
        """Tests that SparkGroupedDataFrameDomain has expected carrier type."""
        self.assertEqual(self.domain.carrier_type, GroupedDataFrame)

    @parameterized.expand(
        [
            (
                pd.DataFrame({"A": [1, 2], "B": ["W", "W"], "C": [4, 5]}),
                pd.DataFrame({"A": [1, 2, 3], "B": ["W", "X", "Y"]}),
                True,
            ),
            (  # Mismatching DataFrame domain (extra column D)
                pd.DataFrame({"A": [1, 2], "B": ["W", "W"], "C": [4, 5], "D": [4, 5]}),
                pd.DataFrame({"A": [1, 2, 3], "B": ["W", "X", "Y"]}),
                False,
            ),
            (  # Mismatching group keys
                pd.DataFrame({"A": [1, 2], "B": ["W", "W"], "C": [4, 5]}),
                pd.DataFrame({"A": [2, 3, 1], "B": ["W", "X", "Y"]}),
                False,
            ),
        ]
    )
    def test_contains(
        self, dataframe: pd.DataFrame, group_keys: pd.DataFrame, expected: bool
    ):
        """Tests that __contains__ works correctly."""
        grouped_data = GroupedDataFrame(
            dataframe=self.spark.createDataFrame(dataframe),
            group_keys=self.spark.createDataFrame(group_keys),
        )
        self.assertEqual(grouped_data in self.domain, expected)

    def test_eq_positive(self):
        """Tests that __eq__ returns True correctly."""
        other_domain = SparkGroupedDataFrameDomain(
            schema=self.schema, group_keys=self.group_keys
        )
        self.assertTrue(self.domain == other_domain)

    def test_eq_negative(self):
        """Tests that __eq__ returns False correctly."""
        mismatching_schema_domain = SparkGroupedDataFrameDomain(
            schema={
                "A": SparkIntegerColumnDescriptor(),
                "B": SparkStringColumnDescriptor(),
                "C": SparkIntegerColumnDescriptor(),
                "D": SparkIntegerColumnDescriptor(),
            },
            group_keys=self.group_keys,
        )
        mismatching_group_keys_domain = SparkGroupedDataFrameDomain(
            schema=self.schema,
            group_keys=self.spark.createDataFrame(
                [(1, "W"), (2, "X"), (3, "Y"), (4, "Z")], schema=["A", "B"]
            ),
        )
        self.assertFalse(self.domain == mismatching_schema_domain)
        self.assertFalse(self.domain == mismatching_group_keys_domain)

    @parameterized.expand(
        [
            (
                {
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                },
                pd.DataFrame({"C": [1, 2]}),
                "Invalid groupby column: {'C'}",
            ),
            (
                {
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                },
                pd.DataFrame({"B": [1, 2]}),
                "Column must be StringType",
                OutOfDomainError,
            ),
        ]
    )
    def test_post_init(
        self,
        schema: SparkColumnsDescriptor,
        group_keys: pd.DataFrame,
        error_msg: str,
        error_type: type = ValueError,
    ):
        """Tests that __post_init__ correctly rejects invalid inputs."""
        with self.assertRaisesRegex(error_type, error_msg):
            SparkGroupedDataFrameDomain(
                schema=schema, group_keys=self.spark.createDataFrame(group_keys)
            )

    def test_post_init_removes_duplicate_keys(self):
        """Tests that __post_init__ removes duplicate group keys."""
        domain = SparkGroupedDataFrameDomain(
            schema=self.schema,
            group_keys=self.spark.createDataFrame(
                [(1, "W"), (1, "W")], schema=["A", "B"]
            ),
        )
        expected = pd.DataFrame({"A": [1], "B": ["W"]})
        actual = domain.group_keys.toPandas()
        self.assert_frame_equal_with_sort(expected, actual)

    def test_eq_no_group_keys(self):
        """Tests that __eq__ works for empty group_keys."""
        domain1 = SparkGroupedDataFrameDomain(
            schema=self.schema,
            group_keys=self.spark.createDataFrame([], schema=StructType([])),
        )
        domain2 = SparkGroupedDataFrameDomain(
            schema=self.schema,
            group_keys=self.spark.createDataFrame([], schema=StructType([])),
        )
        self.assertTrue(domain1 == domain2)

    def test_validate_with_nulls(self):
        """Test that validate works correctly with nulls."""
        group_keys_with_nulls = self.spark.createDataFrame(
            [(1, "W"), (2, "X"), (3, None)], schema=["A", "B"]
        )
        group_keys_without_nulls = self.spark.createDataFrame(
            [(1, "W"), (1, "X")], schema=["A", "B"]
        )
        dataframe_with_nulls = self.spark.createDataFrame(
            [(1, "W", 0), (None, "X", 1), (None, "X", 2)], schema=["A", "B", "C"]
        )
        domain_with_nulls = SparkGroupedDataFrameDomain(
            schema=self.schema, group_keys=group_keys_with_nulls
        )
        # everything matches, no exception should be raised
        domain_with_nulls.validate(
            GroupedDataFrame(
                dataframe=dataframe_with_nulls, group_keys=group_keys_with_nulls
            )
        )
        # group keys don't match, should raise exception
        with self.assertRaises(OutOfDomainError):
            domain_with_nulls.validate(
                GroupedDataFrame(
                    dataframe=dataframe_with_nulls, group_keys=group_keys_without_nulls
                )
            )
        domain_without_nulls = SparkGroupedDataFrameDomain(
            schema=self.schema, group_keys=group_keys_without_nulls
        )
        # group keys match, nulls should be fine
        domain_without_nulls.validate(
            GroupedDataFrame(
                dataframe=dataframe_with_nulls, group_keys=group_keys_without_nulls
            )
        )
        # group keys don't match, should raise exception
        with self.assertRaises(OutOfDomainError):
            domain_without_nulls.validate(
                GroupedDataFrame(
                    dataframe=dataframe_with_nulls, group_keys=group_keys_with_nulls
                )
            )

    def test_eq_with_nulls(self):
        """Tests that eq works correctly with null values."""
        domain_with_nulls1 = SparkGroupedDataFrameDomain(
            schema=self.schema,
            group_keys=self.spark.createDataFrame(
                [(1, "W"), (1, "X"), (None, "X")], schema=["A", "B"]
            ),
        )
        domain_with_nulls2 = SparkGroupedDataFrameDomain(
            schema=self.schema,
            group_keys=self.spark.createDataFrame(
                [(1, "W"), (1, "X"), (None, "X")], schema=["A", "B"]
            ),
        )
        self.assertEqual(domain_with_nulls1, domain_with_nulls2)
        domain_without_nulls = SparkGroupedDataFrameDomain(
            schema=self.schema,
            group_keys=self.spark.createDataFrame(
                [(1, "W"), (1, "X")], schema=["A", "B"]
            ),
        )
        self.assertNotEqual(domain_with_nulls1, domain_without_nulls)

    def test_repr(self):
        """Tests that __repr__ works correctly."""
        expected = (
            "SparkGroupedDataFrameDomain(schema={'A': SparkIntegerColumnDescriptor("
            "allow_null=True, size=64), 'B': SparkStringColumnDescriptor(allow_null="
            "True), 'C': SparkIntegerColumnDescriptor(allow_null=True, size=64)},"
            " group_keys=DataFrame[A: bigint, B: string])"
        )
        self.assertEqual(repr(self.domain), expected)


class TestSparkColumnDescriptors(PySparkTest):
    r"""Tests for subclasses of class SparkColumnDescriptor.

    See subclasses of
    :class:`~tmlt.core.domains.spark_domains.SparkColumnDescriptor`\ s."""

    def setUp(self):
        """Setup"""
        self.int32_column_descriptor = SparkIntegerColumnDescriptor(size=32)
        self.int64_column_descriptor = SparkIntegerColumnDescriptor(size=64)
        self.float32_column_descriptor = SparkFloatColumnDescriptor(size=32)
        self.str_column_descriptor = SparkStringColumnDescriptor(allow_null=False)
        self.date_column_descriptor = SparkDateColumnDescriptor()
        self.timestamp_column_descriptor = SparkTimestampColumnDescriptor()
        self.test_df = self.spark.createDataFrame(
            [
                (
                    1,
                    2,
                    1.0,
                    "X",
                    datetime.date.fromisoformat("1970-01-01"),
                    datetime.datetime.fromisoformat("1970-01-01 00:00:00.000+00:00"),
                ),
                (
                    11,
                    239,
                    2.0,
                    None,
                    datetime.date.fromisoformat("2022-01-01"),
                    datetime.datetime.fromisoformat("2022-01-01 08:30:00.000+00:00"),
                ),
            ],
            schema=StructType(
                [
                    StructField("A", IntegerType(), False),
                    StructField("B", LongType(), False),
                    StructField("C", FloatType(), False),
                    StructField("D", StringType(), True),
                    StructField("E", DateType(), True),
                    StructField("F", TimestampType(), True),
                ]
            ),
        )

    @parameterized.expand(
        [
            (SparkIntegerColumnDescriptor(size=32), NumpyIntegerDomain(size=32)),
            (SparkIntegerColumnDescriptor(size=64), NumpyIntegerDomain(size=64)),
            (SparkFloatColumnDescriptor(size=64), NumpyFloatDomain(size=64)),
            (
                SparkFloatColumnDescriptor(size=64, allow_inf=True),
                NumpyFloatDomain(size=64, allow_inf=True),
            ),
            (
                SparkFloatColumnDescriptor(size=64, allow_nan=True),
                NumpyFloatDomain(size=64, allow_nan=True),
            ),
            (
                SparkStringColumnDescriptor(allow_null=True),
                NumpyStringDomain(allow_null=True),
            ),
            (
                SparkStringColumnDescriptor(allow_null=False),
                NumpyStringDomain(allow_null=False),
            ),
        ]
    )
    def test_to_numpy_domain(
        self, descriptor: SparkColumnDescriptor, expected_domain: Domain
    ):
        """Tests that to_numpy_domain works correctly."""
        self.assertEqual(descriptor.to_numpy_domain(), expected_domain)

    @parameterized.expand(
        [
            (
                SparkIntegerColumnDescriptor(allow_null=True),
                "Nullable column does not have corresponding NumPy domain.",
            ),
            (
                SparkDateColumnDescriptor(),
                "NumPy does not have support for date types.",
            ),
            (
                SparkTimestampColumnDescriptor(),
                "NumPy does not have support for timestamp types.",
            ),
        ]
    )
    def test_to_numpy_domain_invalid(
        self, descriptor: SparkColumnDescriptor, expected_error: str
    ):
        """Tests that to_numpy_domain raises appropriate exceptions."""
        with self.assertRaisesRegex(RuntimeError, expected_error):
            descriptor.to_numpy_domain()

    @parameterized.expand(
        [
            ("A", "int32"),
            ("B", "int64"),
            ("C", "float32"),
            ("D", "str"),
            ("E", "date"),
            ("F", "timestamp"),
        ]
    )
    def test_validate_column(self, col_name: str, col_type: str):
        """Tests that validate_column works correctly."""
        if col_type == "int32":
            self.int32_column_descriptor.validate_column(self.test_df, col_name)
        else:
            with self.assertRaisesRegex(
                OutOfDomainError,
                r"Column must be IntegerType(\(\))?, instead it is "
                f"{self.test_df.schema[col_name].dataType}.",
            ):
                self.int32_column_descriptor.validate_column(self.test_df, col_name)

        if col_type == "int64":
            self.int64_column_descriptor.validate_column(self.test_df, col_name)
        else:
            with self.assertRaisesRegex(
                OutOfDomainError,
                r"Column must be LongType(\(\))?, instead it is "
                f"{self.test_df.schema[col_name].dataType}.",
            ):
                self.int64_column_descriptor.validate_column(self.test_df, col_name)

        if col_type == "float32":
            self.float32_column_descriptor.validate_column(self.test_df, col_name)
        else:
            with self.assertRaisesRegex(
                OutOfDomainError,
                r"Column must be FloatType(\(\))?, instead it is "
                f"{self.test_df.schema[col_name].dataType}.",
            ):
                self.float32_column_descriptor.validate_column(self.test_df, col_name)

        if col_type == "str":
            with self.assertRaisesRegex(
                OutOfDomainError, "Column contains null values."
            ):
                self.str_column_descriptor.validate_column(self.test_df, col_name)
        else:
            with self.assertRaisesRegex(
                OutOfDomainError,
                r"Column must be StringType(\(\))?, instead it is "
                f"{self.test_df.schema[col_name].dataType}.",
            ):
                self.str_column_descriptor.validate_column(self.test_df, col_name)

        if col_type == "date":
            self.date_column_descriptor.validate_column(self.test_df, col_name)
        else:
            with self.assertRaisesRegex(
                OutOfDomainError,
                r"Column must be DateType(\(\))?, instead it is "
                f"{self.test_df.schema[col_name].dataType}.",
            ):
                self.date_column_descriptor.validate_column(self.test_df, col_name)

        if col_type == "timestamp":
            self.timestamp_column_descriptor.validate_column(self.test_df, col_name)
        else:
            with self.assertRaisesRegex(
                OutOfDomainError,
                r"Column must be TimestampType(\(\))?, instead it is "
                f"{self.test_df.schema[col_name].dataType}.",
            ):
                self.timestamp_column_descriptor.validate_column(self.test_df, col_name)

    @parameterized.expand(
        [
            (SparkIntegerColumnDescriptor(size=32), "int32"),
            (SparkIntegerColumnDescriptor(allow_null=True, size=32), "int32_null"),
            (SparkIntegerColumnDescriptor(), "int64"),
            (SparkIntegerColumnDescriptor(allow_null=True), "int64_null"),
            (SparkFloatColumnDescriptor(), "float64"),
            (SparkFloatColumnDescriptor(size=32), "float32"),
            (SparkFloatColumnDescriptor(size=32, allow_nan=True), "float32_nan"),
            (SparkStringColumnDescriptor(), "str"),
            (SparkStringColumnDescriptor(allow_null=True), "str_null"),
            (SparkDateColumnDescriptor(), "date"),
            (SparkDateColumnDescriptor(allow_null=True), "date_null"),
            (SparkTimestampColumnDescriptor(), "timestamp"),
            (SparkTimestampColumnDescriptor(allow_null=True), "timestamp_null"),
        ]
    )
    def test_eq(self, candidate: Any, col_type: str):
        """Tests that __eq__ works correctly."""
        self.assertEqual(self.int32_column_descriptor == candidate, col_type == "int32")
        self.assertEqual(self.int64_column_descriptor == candidate, col_type == "int64")
        self.assertEqual(
            self.float32_column_descriptor == candidate, col_type == "float32"
        )
        self.assertEqual(self.str_column_descriptor == candidate, col_type == "str")
        self.assertEqual(self.date_column_descriptor == candidate, col_type == "date")
        self.assertEqual(
            self.timestamp_column_descriptor == candidate, col_type == "timestamp"
        )



"""Tests for :mod:`~tmlt.core.utils.truncation`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2022

import itertools
from typing import List, Tuple

import pandas as pd
from parameterized import parameterized

from tmlt.core.utils.testing import PySparkTest
from tmlt.core.utils.truncation import drop_large_groups, truncate_large_groups


class TestTruncateLargeGroups(PySparkTest):
    """Tests for :meth:`~tmlt.core.utils.truncation.truncate_large_groups`."""

    @parameterized.expand(
        [
            (2, [(1, "x"), (1, "y"), (1, "z"), (1, "w")], 2),
            (2, [(1, "x")], 1),
            (0, [(1, "x"), (1, "y"), (1, "z"), (1, "w")], 0),
        ]
    )
    def test_correctness(self, threshold: int, rows: List[Tuple], expected_count: int):
        """Tests that truncate_large_groups works correctly."""
        df = self.spark.createDataFrame(rows, schema=["A", "B"])
        self.assertEqual(
            truncate_large_groups(df, ["A"], threshold).count(), expected_count
        )

    def test_consistency(self):
        """Tests that truncate_large_groups does not truncate randomly across calls."""
        df = self.spark.createDataFrame([(i,) for i in range(1000)], schema=["A"])

        expected_output = truncate_large_groups(df, ["A"], 5).toPandas()
        for _ in range(5):
            self.assert_frame_equal_with_sort(
                truncate_large_groups(df, ["A"], 5).toPandas(), expected_output
            )

    def test_rows_dropped_consistently(self):
        """Tests that truncate_large_groups drops that same rows for unchanged keys."""
        df1 = self.spark.createDataFrame(
            [("A", 1), ("B", 2), ("B", 3)], schema=["W", "X"]
        )
        df2 = self.spark.createDataFrame(
            [("A", 0), ("A", 1), ("B", 2), ("B", 3)], schema=["W", "X"]
        )

        df1_truncated = truncate_large_groups(df1, ["W"], 1)
        df2_truncated = truncate_large_groups(df2, ["W"], 1)
        self.assert_frame_equal_with_sort(
            df1_truncated.filter("W='B'").toPandas(),
            df2_truncated.filter("W='B'").toPandas(),
        )

    def test_hash_truncation_order_agnostic(self):
        """Tests that truncate_large_groups doesn't depend on row order."""
        df_rows = [(1, 2, "A"), (3, 4, "A"), (5, 6, "A"), (7, 8, "B")]

        truncated_dfs: List[pd.DataFrame] = []
        for permutation in itertools.permutations(df_rows, 4):
            df = self.spark.createDataFrame(list(permutation), schema=["W", "X", "Y"])
            truncated_dfs.append(truncate_large_groups(df, ["Y"], 1).toPandas())
        for df in truncated_dfs[1:]:
            self.assert_frame_equal_with_sort(first_df=truncated_dfs[0], second_df=df)


class TestDropLargeGroups(PySparkTest):
    """Tests for :meth:`~tmlt.core.utils.truncation.drop_large_groups`."""

    @parameterized.expand(
        [
            (1, [(1, "A"), (1, "B"), (2, "C")], [(2, "C")]),
            (1, [(1, "A"), (2, "C")], [(1, "A"), (2, "C")]),
            (2, [(1, "A"), (2, "C"), (2, "D"), (2, "E")], [(1, "A")]),
            (1, [(1, "A"), (1, "B"), (2, "C"), (2, "D"), (2, "E")], []),
            (0, [(1, "x"), (2, "y"), (3, "z"), (3, "w")], []),
        ]
    )
    def test_correctness(
        self, threshold: int, input_rows: List[Tuple], expected: List[Tuple]
    ):
        """Tests that drop_large_groups works correctly."""
        df = self.spark.createDataFrame(input_rows, schema=["A", "B"])
        actual = drop_large_groups(df, ["A"], threshold).toPandas()
        expected = pd.DataFrame.from_records(expected, columns=["A", "B"])
        self.assert_frame_equal_with_sort(actual, expected)



"""Probability functions for distributions commonly used in differential privacy."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2022

from functools import lru_cache
from typing import overload

import numpy as np
import sympy as sp

from tmlt.core.utils.exact_number import ExactNumber, ExactNumberInput


@overload
def double_sided_geometric_pmf(k: int, alpha: float) -> float:
    ...


@overload
def double_sided_geometric_pmf(k: np.ndarray, alpha: float) -> np.ndarray:
    ...


@overload
def double_sided_geometric_pmf(k: int, alpha: np.ndarray) -> np.ndarray:
    ...


def double_sided_geometric_pmf(
    k, alpha
):  # pylint: disable=missing-type-doc, missing-return-type-doc
    r"""Returns the pmf for a double-sided geometric distribution at k.

    For :math:`k \in \mathbb{Z}`

    .. math::

        f(k)=
        \frac
            {e^{1 / \alpha} - 1}
            {e^{1 / \alpha} + 1}
        \cdot
        e^{\frac{-\mid k \mid}{\alpha}}

    A double sided geometric distribution is the difference between two geometric
    distributions (It can be sampled by sampling two values from a geometric
    distribution, and taking their difference).

    See section 4.1 in :cite:`BalcerV18` or scipy.stats.geom for more information (Note
    that the parameter :math:`p` used in scipy.stats.geom is related to :math:`\alpha`
    through :math:`p = 1 - e^{-1 / \alpha}`).

    Args:
        k: The values to calculate the pmf for.
        alpha: The scale of the geometric distribution.
    """
    return (
        (np.exp(1 / alpha) - 1) / (np.exp(1 / alpha) + 1) * np.exp(-np.abs(k) / alpha)
    )


@overload
def double_sided_geometric_cmf(k: int, alpha: float) -> float:
    ...


@overload
def double_sided_geometric_cmf(k: np.ndarray, alpha: float) -> np.ndarray:
    ...


def double_sided_geometric_cmf(
    k, alpha
):  # pylint: disable=missing-type-doc, missing-return-type-doc
    r"""Returns the cmf for a double-sided geometric distribution at k.

    For :math:`k \in \mathbb{Z}`

    .. math::

        F(k) = \begin{cases}
        \frac{e^{1 / \alpha}}{e^{1 / \alpha} + 1} \cdot e^\frac{k}{\alpha} &
        \text{if k} \le 0 \\
        1 - \frac{1}{e^{1 / \alpha} + 1}\cdot e^{\frac{-k}{\alpha}} &
        \text{otherwise} \\
        \end{cases}

    See :func:`double_sided_geometric_pmf` for more information.

    Args:
        k: The values to calculate the pmf for.
        alpha: The scale of the geometric distribution.
    """
    if isinstance(k, int):
        return double_sided_geometric_cmf(np.array([k]), alpha)[0]
    return np.where(
        k <= 0,
        np.exp(1 / alpha) / (np.exp(1 / alpha) + 1) * np.exp(k / alpha),
        1 - np.exp(-k / alpha) / (np.exp(1 / alpha) + 1),
    )


def double_sided_geometric_cmf_exact(
    k: ExactNumberInput, alpha: ExactNumberInput
) -> ExactNumber:
    """Returns exact value of the cmf for a double-sided geometric distribution at k.

    See :func:`double_sided_geometric_cmf` for more information.

    Args:
        k: The values to calculate the pmf for.
        alpha: The scale of the geometric distribution.
    """
    k_expr = ExactNumber(k).expr
    alpha_expr = ExactNumber(alpha).expr
    if k_expr <= 0:
        p = (
            sp.exp(1 / alpha_expr)
            / (sp.exp(1 / alpha_expr) + 1)
            * sp.exp(k_expr / alpha_expr)
        )
    else:
        p = 1 - 1 / (sp.exp(k_expr / alpha_expr) * (sp.exp(1 / alpha_expr) + 1))
    return ExactNumber(p)


@overload
def double_sided_geometric_inverse_cmf(p: float, alpha: float) -> int:
    ...


@overload
def double_sided_geometric_inverse_cmf(p: np.ndarray, alpha: float) -> np.ndarray:
    ...


def double_sided_geometric_inverse_cmf(
    p, alpha
):  # pylint: disable=missing-type-doc, missing-return-type-doc
    r"""Returns the inverse cmf of a double-sided geometric distribution at p.

    In other words, it returns the smallest k s.t. CMF(k) >= p.

    For :math:`p \in [0, 1]`

    .. math::

        F(k) = \begin{cases}
        \alpha\ln(p\cdot\frac{e^{\frac{1}{\alpha}} + 1}{e^{\frac{1}{\alpha}}}) &
        \text{if p} \le 0.5 \\
        -\alpha\ln((e^{\frac{1}{\alpha}} + 1)(1 - p)) &
        \text{otherwise} \\
        \end{cases}

    See :func:`double_sided_geometric_pmf` for more information.
    """
    if isinstance(p, float):
        return double_sided_geometric_inverse_cmf(np.array([p]), alpha)[0]
    return np.where(
        p <= 0.5,
        np.ceil(alpha * np.log(p * (np.exp(1 / alpha) + 1) / np.exp(1 / alpha))),
        np.ceil(-alpha * np.log((np.exp(1 / alpha) + 1) * (1 - p))),
    )


@lru_cache(None)
def _discrete_gaussian_normalizing_constant(sigma_squared: float):
    """Returns the normalizing factor for :func:`discrete_gaussian_pmf`."""
    # This method for calculating the upper/lower bound to consider is used in
    # https://github.com/IBM/discrete-gaussian-differential-privacy/blob/master/testing-kolmogorov-discretegaussian.py
    # They say
    # "Compute the normalizing constant from -bounds to bounds instead of -inf to inf:
    # negligible error, as bounds is at least 50 standard deviations from 0,
    # and the tails decay exponentially."
    bound = max(10000, int(sigma_squared * 50))
    # +1 is the contribution from y = 0
    # The factor of 2 is because y/-y contribute the same for y != 0
    return 1 + 2 * np.sum(np.exp(-np.arange(1, bound) ** 2 / (2 * sigma_squared)))


@overload
def discrete_gaussian_pmf(k: int, sigma_squared: float) -> float:
    ...


@overload
def discrete_gaussian_pmf(k: np.ndarray, sigma_squared: float) -> np.ndarray:
    ...


def discrete_gaussian_pmf(k, sigma_squared):
    r"""Returns the pmf for a discrete gaussian distribution at k.

    For :math:`k \in \mathbb{Z}`

    .. math::
        :label: discrete_gaussian_pmf

        f(k) = \frac
        {e^{-k^2/2\sigma^2}}
        {
            \sum_{n\in \mathbb{Z}}
            e^{-n^2/2\sigma^2}
        }

    See :cite:`Canonne0S20` for more information. The formula above is based on
    Definition 1.

    The implementation also referenced the paper's implementation, which can be found at
    https://github.com/IBM/discrete-gaussian-differential-privacy.
    """
    return np.exp(
        -(k ** 2) / (2 * sigma_squared)
    ) / _discrete_gaussian_normalizing_constant(sigma_squared)


@overload
def discrete_gaussian_cmf(k: int, sigma_squared: float) -> float:
    ...


@overload
def discrete_gaussian_cmf(k: np.ndarray, sigma_squared: float) -> np.ndarray:
    ...


def discrete_gaussian_cmf(k, sigma_squared):
    """Returns the cmf for a discrete gaussian distribution at k.

    See :eq:`discrete_gaussian_pmf` for the probability mass function.
    """
    if isinstance(k, int):
        return discrete_gaussian_cmf(np.array([k]), sigma_squared)[0]
    # _discrete_gaussian_normalizing_constant explains calculating the bound this way
    lower_bound = min(-10000, -int(sigma_squared * 50), np.min(k))
    unnormalized_cmf = np.cumsum(
        np.exp(
            -(np.abs(np.arange(lower_bound, np.max(k) + 1) ** 2)) / (2 * sigma_squared)
        )
    )
    return unnormalized_cmf[k - lower_bound] / _discrete_gaussian_normalizing_constant(
        sigma_squared
    )


@overload
def discrete_gaussian_inverse_cmf(p: float, sigma_squared: float) -> int:
    ...


@overload
def discrete_gaussian_inverse_cmf(p: np.ndarray, sigma_squared: float) -> np.ndarray:
    ...


def discrete_gaussian_inverse_cmf(p, sigma_squared):
    """Returns the inverse cmf for a discrete gaussian distribution at p.

    In other words, it returns the smallest k s.t. CMF(k) >= p.

    This is not terribly performant. Don't use it in places where it's
    going to get called a lot.
    """
    if isinstance(p, np.ndarray):
        return np.vectorize(discrete_gaussian_inverse_cmf)(p, sigma_squared)

    unnormalized_cmf = p * _discrete_gaussian_normalizing_constant(sigma_squared)

    def unnormalized_pmf(k):
        return np.exp(-(k ** 2) / (2 * sigma_squared))

    # _discrete_gaussian_normalizing_constant explains calculating the bound this way
    lower_bound = min(-10000, -int(sigma_squared * 50))

    # if k < lower bound, cmf(k) returns pmf(k), so we can binary search for k.
    if unnormalized_pmf(lower_bound) > unnormalized_cmf:
        hi = lower_bound
        lo = lower_bound * 2
        while unnormalized_pmf(lo) > unnormalized_cmf:
            hi = lo
            lo *= 2
        while hi != lo:
            mid = hi / 2 + lo / 2
            if unnormalized_pmf(mid) == unnormalized_cmf:
                return mid
            elif unnormalized_pmf(mid) < unnormalized_cmf:
                lo = mid
            else:
                hi = mid
        return hi

    remaining_cmf = unnormalized_cmf
    current_step = lower_bound
    while remaining_cmf > 0:
        remaining_cmf -= unnormalized_pmf(current_step)
        current_step += 1

    return current_step - 1



"""Unit tests for :mod:`~tmlt.core.domains.pandas_domains`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2022
from typing import Any, Optional
from unittest.case import TestCase

import numpy as np
import pandas as pd
from parameterized import parameterized

from tmlt.core.domains.base import Domain, OutOfDomainError
from tmlt.core.domains.numpy_domains import (
    NumpyFloatDomain,
    NumpyIntegerDomain,
    NumpyStringDomain,
)
from tmlt.core.domains.pandas_domains import PandasDataFrameDomain, PandasSeriesDomain
from tmlt.core.utils.testing import assert_property_immutability, get_all_props


class TestPandasSeriesDomain(TestCase):
    """Tests for :class:`~tmlt.core.domains.pandas_domains.PandasSeriesDomain`."""

    def setUp(self):
        """Setup."""
        self.int32_series_domain = PandasSeriesDomain(
            element_domain=NumpyIntegerDomain(size=32)
        )
        self.int64_series_domain = PandasSeriesDomain(
            element_domain=NumpyIntegerDomain(size=64)
        )
        self.float32_inf_series_domain = PandasSeriesDomain(
            element_domain=NumpyFloatDomain(size=32, allow_inf=True)
        )

    @parameterized.expand(
        [
            (
                PandasSeriesDomain(element_domain=NumpyIntegerDomain()),
                False,
                True,
                False,
            ),
            (
                PandasSeriesDomain(element_domain=NumpyIntegerDomain(size=32)),
                True,
                False,
                False,
            ),
            (
                PandasSeriesDomain(element_domain=NumpyFloatDomain()),
                False,
                False,
                False,
            ),
            (
                PandasSeriesDomain(element_domain=NumpyFloatDomain(size=32)),
                False,
                False,
                False,
            ),
            (
                PandasSeriesDomain(
                    element_domain=NumpyFloatDomain(size=32, allow_inf=True)
                ),
                False,
                False,
                True,
            ),
            (NumpyFloatDomain(size=32), False, False, False),
        ]
    )
    def test_eq(
        self,
        domain: Domain,
        int32_series: bool,
        int64_series: bool,
        float32_series: bool,
    ):
        """Tests that __eq__ works correctly."""
        self.assertEqual(domain == self.int32_series_domain, int32_series)
        self.assertEqual(domain == self.int64_series_domain, int64_series)
        self.assertEqual(domain == self.float32_inf_series_domain, float32_series)

    @parameterized.expand(
        [
            (pd.Series([1, 2], dtype=np.dtype("int32")), True, False, False),
            (pd.Series([1, 2], dtype=np.dtype("int64")), False, True, False),
            (pd.Series([1.0, 2.0], dtype=np.dtype("float32")), False, False, True),
            (
                pd.Series([1.0, float("inf")], dtype=np.dtype("float32")),
                False,
                False,
                True,
            ),
            (
                pd.Series([1.0, float("nan")], dtype=np.dtype("float32")),
                False,
                False,
                False,
            ),
            ("Not a member", False, False, False),
        ]
    )
    def test_validate(
        self,
        candidate: Any,
        in_int32_domain: bool,
        in_int64_domain: bool,
        in_float32_domain: bool,
    ):
        """Tests that validate works correctly."""
        if not isinstance(candidate, pd.Series):
            self.assertRaisesRegex(
                OutOfDomainError,
                f"Value must be {pd.Series}, instead it is {candidate.__class__}",
            )
            return

        exception = "Found invalid value in Series: *"
        if in_int32_domain:
            self.assertEqual(self.int32_series_domain.validate(candidate), None)
        else:
            with self.assertRaisesRegex(OutOfDomainError, exception):
                self.int32_series_domain.validate(candidate)

        if in_int64_domain:
            self.assertEqual(self.int64_series_domain.validate(candidate), None)
        else:
            with self.assertRaisesRegex(OutOfDomainError, exception):
                self.int64_series_domain.validate(candidate)

        if in_float32_domain:
            self.assertEqual(self.float32_inf_series_domain.validate(candidate), None)
        else:
            with self.assertRaisesRegex(OutOfDomainError, exception):
                self.float32_inf_series_domain.validate(candidate)


class TestPandasDataFrameDomain(TestCase):
    """Tests for class PandasDataFrameDomain.

    Tests :class:`~tmlt.core.domains.pandas_domains.PandasDataFrameDomain`."""

    def setUp(self):
        """Create a test PandasDataFrameDomain"""
        self.pdfd = PandasDataFrameDomain(
            {
                "A": PandasSeriesDomain(NumpyIntegerDomain()),
                "B": PandasSeriesDomain(NumpyFloatDomain(allow_inf=True)),
                "C": PandasSeriesDomain(NumpyStringDomain()),
            }
        )

    def test_constructor_mutable_arguments(self):
        """Tests that mutable constructor arguments are copied."""
        schema = {"A": PandasSeriesDomain(NumpyIntegerDomain())}
        domain = PandasDataFrameDomain(schema=schema)
        schema["A"] = NumpyFloatDomain()
        self.assertDictEqual(
            domain.schema, {"A": PandasSeriesDomain(NumpyIntegerDomain())}
        )

    @parameterized.expand(get_all_props(PandasDataFrameDomain))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        assert_property_immutability(self.pdfd, prop_name)

    def test_bad_init(self):
        """Test that PandasDataFrameDomain raises error when create with wrong type."""
        with self.assertRaises(TypeError):
            PandasDataFrameDomain(schema="incorrect type")

    @parameterized.expand(
        [
            (pd.DataFrame({"A": [1, 2], "B": [1.0, 2.0], "C": ["1", "2"]}), None),
            # wrong column type
            (
                pd.DataFrame({"A": [1, 2], "B": [1, 2], "C": ["1", "2"]}),
                "Found invalid value in column 'B': Found invalid value in Series: "
                f"Value must be {np.float64}, instead it is {np.int64}.",
            ),
            # missing column
            (
                pd.DataFrame({"A": [1, 2], "B": [1.0, 2.0]}),
                "Columns are not as expected. DataFrame and Domain must contain the "
                r"same columns in the same order.\nDataFrame columns: \['A', 'B'\]\n"
                r"Domain columns: \['A', 'B', 'C'\]",
            ),
            # columns out of order
            (
                pd.DataFrame({"A": [1, 2], "C": ["1", "2"], "B": [1.0, 2.0]}),
                "Columns are not as expected. DataFrame and Domain must contain the "
                "same columns in the same order.\nDataFrame columns: "
                r"\['A', 'C', 'B'\]\n"
                r"Domain columns: \['A', 'B', 'C'\]",
            ),
            # wrong type
            (
                pd.Series([1, 2, 3]),
                f"Value must be {pd.DataFrame}, instead it is {pd.Series}.",
            ),
            # duplicated columns
            (
                pd.DataFrame(
                    [["A", "A", "B", 1.1], ["V", "V", "E", 1.2], ["A", "A", "V", 1.3]],
                    columns=["A", "A", "B", "C"],
                ),
                r"Some columns are duplicated, \['A'\]",
            ),
        ]
    )
    def test_validate(self, value: Any, exception: Optional[str]):
        """Test that validate works as expected."""
        if exception is not None:
            with self.assertRaisesRegex(OutOfDomainError, exception):
                self.pdfd.validate(value)
        else:
            self.assertEqual(self.pdfd.validate(value), exception)

    @parameterized.expand(
        [
            (
                PandasDataFrameDomain(
                    {
                        "A": PandasSeriesDomain(NumpyIntegerDomain()),
                        "B": PandasSeriesDomain(NumpyFloatDomain(allow_inf=True)),
                        "C": PandasSeriesDomain(NumpyStringDomain()),
                    }
                ),
                True,
            ),
            # wrong column type
            (
                PandasDataFrameDomain(
                    {
                        "A": PandasSeriesDomain(NumpyIntegerDomain(size=32)),
                        "B": PandasSeriesDomain(NumpyFloatDomain(allow_inf=True)),
                        "C": PandasSeriesDomain(NumpyStringDomain()),
                    }
                ),
                False,
            ),
            # columns out of order
            (
                PandasDataFrameDomain(
                    {
                        "A": PandasSeriesDomain(NumpyIntegerDomain()),
                        "C": PandasSeriesDomain(NumpyStringDomain()),
                        "B": PandasSeriesDomain(NumpyFloatDomain(allow_inf=True)),
                    }
                ),
                False,
            ),
        ]
    )
    def test_eq(self, other: Any, is_match: bool):
        """Test that to_spark_domain works as expected."""
        self.assertEqual(self.pdfd == other, is_match)

    def test_repr(self):
        """Tests that __repr__ works correctly"""
        domain = PandasDataFrameDomain({"A": PandasSeriesDomain(NumpyIntegerDomain())})

        expected = (
            "PandasDataFrameDomain(schema={'A':"
            " PandasSeriesDomain(element_domain=NumpyIntegerDomain(size=64))})"
        )
        self.assertEqual(repr(domain), expected)



"""Wrappers for changing a measurements's output measure."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2022

from typing import Any, Tuple

import sympy as sp
from typeguard import typechecked

from tmlt.core.measurements.base import Measurement
from tmlt.core.measures import ApproxDP, PureDP, RhoZCDP
from tmlt.core.utils.exact_number import ExactNumber, ExactNumberInput


class PureDPToRhoZCDP(Measurement):
    """Measurement for converting pure DP to zCDP."""

    @typechecked
    def __init__(self, pure_dp_measurement: Measurement):
        """Constructor.

        Args:
            pure_dp_measurement: The pure DP measurement to convert.
        """
        if pure_dp_measurement.output_measure != PureDP():
            raise ValueError("Input measure must be pure dp.")
        if pure_dp_measurement.is_interactive:
            raise ValueError("Can only convert non-interactive measurements.")

        self._pure_dp_measurement = pure_dp_measurement

        super().__init__(
            input_domain=pure_dp_measurement.input_domain,
            input_metric=pure_dp_measurement.input_metric,
            output_measure=RhoZCDP(),
            is_interactive=False,
        )

    @property
    def pure_dp_measurement(self) -> Measurement:
        """Return the wrapped pure DP measurement."""
        return self._pure_dp_measurement

    @typechecked
    def privacy_function(self, d_in: Any) -> ExactNumber:
        r"""Returns the smallest d_out satisfied by the measurement.

        The returned d_out (:math:`\rho`) is :math:`\rho=\frac{\epsilon^2}{2}`

        where :math:`\epsilon` is the d_out returned by :attr:`~.pure_dp_measurement`'s
        :meth:`~.Measurement.privacy_function` on d_in.

        Args:
            d_in: Distance between inputs under input_metric.

        Raises:
            NotImplementedError: If self.pure_dp_measurement.privacy_function(d_in)
                raises :class:`NotImplementedError`.
        """
        epsilon = self.pure_dp_measurement.privacy_function(d_in)
        assert isinstance(epsilon, ExactNumber)
        rho = (epsilon ** 2) / 2
        self.output_measure.validate(rho)
        return rho

    @typechecked
    def privacy_relation(self, d_in: Any, d_out: ExactNumberInput) -> bool:
        r"""Return True if close inputs produce close outputs.

        Let :math:`\epsilon = \sqrt{2 \cdot \rho}`

        Returns self.pure_dp_measurement.privacy_relation(d_in, :math:`\epsilon`)

        where :math:`\rho` is the input argument "d_out".

        Args:
            d_in: Distance between inputs under input_metric.
            d_out: Distance between outputs under output_measure.
        """
        self.input_metric.validate(d_in)
        self.output_measure.validate(d_out)
        pure_dp_epsilon = sp.sqrt(2 * ExactNumber(d_out).expr)
        return self.pure_dp_measurement.privacy_relation(d_in, pure_dp_epsilon)

    def __call__(self, data: Any):
        """Apply measurement."""
        return self.pure_dp_measurement(data)


class PureDPToApproxDP(Measurement):
    """Measurement for converting pure DP to approximate DP."""

    @typechecked
    def __init__(self, pure_dp_measurement: Measurement):
        """Constructor.

        Args:
            pure_dp_measurement: The pure DP measurement to convert.
        """
        if pure_dp_measurement.output_measure != PureDP():
            raise ValueError("Input measure must be pure DP.")
        if pure_dp_measurement.is_interactive:
            raise ValueError("Can only convert non-interactive measurements.")

        self._pure_dp_measurement = pure_dp_measurement

        super().__init__(
            input_domain=pure_dp_measurement.input_domain,
            input_metric=pure_dp_measurement.input_metric,
            output_measure=ApproxDP(),
            is_interactive=False,
        )

    @property
    def pure_dp_measurement(self) -> Measurement:
        """Return the wrapped pure DP measurement."""
        return self._pure_dp_measurement

    @typechecked
    def privacy_function(self, d_in: Any) -> Tuple[ExactNumber, ExactNumber]:
        r"""Returns the smallest d_out satisfied by the measurement.

        Returns (self.pure_dp_measurement.privacy_function(d_in), 0).

        Every (:math:`\epsilon`)-DP measurement is also (:math:`\epsilon`, 0)-DP.

        Args:
            d_in: Distance between inputs under input_metric.

        Raises:
            NotImplementedError: If :attr:`~.pure_dp_measurement`'s
                :meth:`~.Measurement.privacy_function` raises
                :class:`NotImplementedError`.
        """
        return self.pure_dp_measurement.privacy_function(d_in), ExactNumber(0)

    @typechecked
    def privacy_relation(
        self, d_in: Any, d_out: Tuple[ExactNumberInput, ExactNumberInput]
    ) -> bool:
        r"""Returns the smallest d_out satisfied by the measurement.

        Returns self.pure_dp_measurement.privacy_relation(d_in, d_out[0])

        Args:
            d_in: Distance between inputs under input_metric.
            d_out: Distance between outputs under output_measure.
        """
        self.input_metric.validate(d_in)
        self.output_measure.validate(d_out)
        return self.pure_dp_measurement.privacy_relation(d_in, d_out[0])

    def __call__(self, data: Any):
        """Apply measurement."""
        return self.pure_dp_measurement(data)


class RhoZCDPToApproxDP(Measurement):
    """Measurement for converting zCDP to approximate DP."""

    @typechecked
    def __init__(self, zcdp_measurement: Measurement):
        """Constructor.

        Args:
            zcdp_measurement: The zCDP measurement to convert.
        """
        if zcdp_measurement.output_measure != RhoZCDP():
            raise ValueError("Input measure must be rho zCDP.")
        if zcdp_measurement.is_interactive:
            raise ValueError("Can only convert non-interactive measurements.")

        self._zcdp_measurement = zcdp_measurement

        super().__init__(
            input_domain=zcdp_measurement.input_domain,
            input_metric=zcdp_measurement.input_metric,
            output_measure=ApproxDP(),
            is_interactive=False,
        )

    @property
    def zcdp_measurement(self) -> Measurement:
        """Return the wrapped zCDP measurement."""
        return self._zcdp_measurement

    @typechecked
    def privacy_relation(
        self, d_in: Any, d_out: Tuple[ExactNumberInput, ExactNumberInput]
    ) -> bool:
        r"""Return True if close inputs produce close outputs.

        Special cases:

        * Every measurement is (:math:`\infty, \delta`)-DP
          for any :math:`\delta \ge 0`
        * Every measurement is (:math:`epsilon, 1`)-DP
          for any :math:`\epsilon \ge 0`
        * :class:`~.RhoZCDP` with :math:`\rho \gt 0` cannot be converted to
          (:math:`\epsilon, 0`)-DP for any finite :math:`\epsilon`

        General case:

        Let :math:`\rho` be the unique solution to
        :math:`\rho - 2 * \sqrt{\rho * log(\frac{1}{\delta})} - \epsilon = 0`

        where:

        * :math:`\delta` is the first element of the input argument "d_out"
        * :math:`\epsilon` is the second element of the input argument "d_out"

        Returns self.zcdp_measurement.privacy_relation(d_in, :math:`\rho`)

        See Proposition 1.3 in :cite:`BunS16` for more information.

        Args:
            d_in: Distance between inputs under input_metric.
            d_out: Distance between outputs under output_measure.
        """
        self.input_metric.validate(d_in)
        self.output_measure.validate(d_out)
        epsilon = ExactNumber(d_out[0]).expr
        delta = ExactNumber(d_out[1]).expr
        if epsilon == sp.oo or delta == 1:
            return True
        if delta == 0:
            return self.zcdp_measurement.privacy_relation(d_in, 0)
        # In the general case, we solve the conversion formula to find rho. It
        # is increasing between rho=0 and rho=infinity, negative in rho=0, and
        # unbounded when rho=infinity, so there is a single solution.
        rho = sp.Symbol("rho")
        solutions = sp.solve(
            rho + sp.Integer(2) * sp.sqrt(rho * sp.log(sp.Integer(1) / delta)) - epsilon
        )
        return self.zcdp_measurement.privacy_relation(d_in, solutions[0])

    def __call__(self, data: Any):
        """Apply measurement."""
        return self.zcdp_measurement(data)



"""Unit tests for :mod:`~tmlt.core.domains.numpy`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2022
from typing import Any, Optional
from unittest.case import TestCase

import numpy as np
from parameterized import parameterized

from tmlt.core.domains.base import Domain, OutOfDomainError
from tmlt.core.domains.numpy_domains import (
    NumpyFloatDomain,
    NumpyIntegerDomain,
    NumpyStringDomain,
)


class TestNumpyFloatDomain(TestCase):
    """Tests for :class:`~tmlt.core.domains.numpy.NumpyFloatDomain`."""

    def setUp(self):
        """Setup."""
        self.float_domain = NumpyFloatDomain()
        self.float32_domain = NumpyFloatDomain(size=32)
        self.nullable_float_domain = NumpyFloatDomain(allow_nan=True)
        self.float_inf_domain = NumpyFloatDomain(allow_inf=True)
        self.nullable_float_inf_domain = NumpyFloatDomain(
            allow_nan=True, allow_inf=True
        )

    @parameterized.expand([(True, None), ("False", True), (True, 0.1)])
    def test_invalid_inputs(self, allow_inf: bool, allow_nan: bool):
        """Test FloatDomain with invalid inputs."""
        with self.assertRaises(TypeError):
            NumpyFloatDomain(allow_inf, allow_nan)

    @parameterized.expand(
        [
            (NumpyFloatDomain(), True, False, False, False, False),
            (NumpyFloatDomain(size=32), False, True, False, False, False),
            (NumpyFloatDomain(allow_nan=True), False, False, True, False, False),
            (NumpyFloatDomain(allow_inf=True), False, False, False, True, False),
            (
                NumpyFloatDomain(allow_nan=True, allow_inf=True),
                False,
                False,
                False,
                False,
                True,
            ),
        ]
    )
    def test_eq(
        self,
        domain: NumpyFloatDomain,
        float_only: bool,
        float32_only: bool,
        nullable_float: bool,
        float_with_inf: bool,
        nullable_float_inf: bool,
    ):
        """Tests that __eq__  works correctly."""
        self.assertEqual(self.float_domain == domain, float_only)
        self.assertEqual(self.float32_domain == domain, float32_only)
        self.assertEqual(self.nullable_float_domain == domain, nullable_float)
        self.assertEqual(self.float_inf_domain == domain, float_with_inf)
        self.assertEqual(self.nullable_float_inf_domain == domain, nullable_float_inf)

    @parameterized.expand(
        [
            (
                np.float64(1.0),
                True,
                False,
                True,
                True,
                True,
                f"Value must be {np.float32}, instead it is {np.float64}.",
            ),
            (
                np.float32(1.0),
                False,
                True,
                False,
                False,
                False,
                f"Value must be {np.float64}, instead it is {np.float32}.",
            ),
            (
                np.float64(float("inf")),
                False,
                False,
                False,
                True,
                True,
                "Value is infinite.",
            ),
            (
                np.float64(-float("inf")),
                False,
                False,
                False,
                True,
                True,
                "Value is infinite.",
            ),
            (
                np.float64(float("nan")),
                False,
                False,
                True,
                False,
                True,
                "Value is NaN.",
            ),
            (
                1,
                False,
                False,
                False,
                False,
                False,
                f"Value must be {np.float64}, instead it is {int}.",
            ),
        ]
    )
    def test_validate(
        self,
        candidate: Any,
        float_only: bool,
        float32_only: bool,
        float_or_nan: bool,
        float_or_inf: bool,
        float_or_nan_or_inf: bool,
        exception: str,
    ):
        """Tests that validate works correctly."""
        if float_only:
            self.assertEqual(self.float_domain.validate(candidate), None)
        else:
            with self.assertRaisesRegex(OutOfDomainError, exception):
                self.float_domain.validate(candidate)

        if float_or_nan:
            self.assertEqual(self.nullable_float_domain.validate(candidate), None)
        else:
            with self.assertRaisesRegex(OutOfDomainError, exception):
                self.nullable_float_domain.validate(candidate)

        if float_or_inf:
            self.assertEqual(self.float_inf_domain.validate(candidate), None)
        else:
            with self.assertRaisesRegex(OutOfDomainError, exception):
                self.float_inf_domain.validate(candidate)

        if float_or_nan_or_inf:
            self.assertEqual(self.nullable_float_inf_domain.validate(candidate), None)
        else:
            with self.assertRaisesRegex(OutOfDomainError, exception):
                self.nullable_float_inf_domain.validate(candidate)

        if float32_only:
            self.assertEqual(self.float32_domain.validate(candidate), None)
        else:
            if isinstance(candidate, int):
                exception = f"Value must be {np.float32}, instead it is {int}."
            else:
                exception = f"Value must be {np.float32}, instead it is {np.float64}."

            with self.assertRaisesRegex(OutOfDomainError, exception):
                self.float32_domain.validate(candidate)


class TestNumpyIntegerDomain(TestCase):
    """Tests for :class:`~tmlt.core.domains.numpy.NumpyIntegerDomain`."""

    def setUp(self):
        """Setup."""
        self.int64_domain = NumpyIntegerDomain()
        self.int32_domain = NumpyIntegerDomain(size=32)

    @parameterized.expand(
        [
            (NumpyFloatDomain(), False, False),
            (NumpyIntegerDomain(), True, False),
            (NumpyIntegerDomain(size=32), False, True),
        ]
    )
    def test_eq(self, domain: Domain, int64_domain: bool, int32_domain: bool):
        """Tests that __eq__  works correctly."""
        self.assertEqual(self.int64_domain == domain, int64_domain)
        self.assertEqual(self.int32_domain == domain, int32_domain)

    @parameterized.expand(
        [
            (np.int64(1), True, False),
            (np.float64(1.1), False, False),
            (np.int32(3), False, True),
        ]
    )
    def test_validate(self, candidate: Any, is_int64: bool, is_int32: bool):
        """Tests that validate works correctly."""
        if is_int64:
            self.assertEqual(self.int64_domain.validate(candidate), None)
        else:
            with self.assertRaisesRegex(
                OutOfDomainError,
                f"Value must be {np.int64}, instead it is {candidate.__class__}",
            ):
                self.int64_domain.validate(candidate)

        if is_int32:
            self.assertEqual(self.int32_domain.validate(candidate), None)
        else:
            with self.assertRaisesRegex(
                OutOfDomainError,
                f"Value must be {np.int32}, instead it is {candidate.__class__}",
            ):
                self.int32_domain.validate(candidate)


class TestNumpyStringDomain(TestCase):
    """Tests for :class:`~tmlt.core.domains.numpy.NumpyStringDomain`."""

    @parameterized.expand(
        [
            (None, NumpyStringDomain(allow_null=True), None),
            (None, NumpyStringDomain(allow_null=False), "Value is null."),
            ("ABC", NumpyStringDomain(allow_null=False), None),
            ("ABC", NumpyStringDomain(allow_null=True), None),
        ]
    )
    def test_validate(self, candidate: Any, domain: Domain, exception: Optional[str]):
        """Tests that validate works correctly."""
        if exception is not None:
            with self.assertRaisesRegex(OutOfDomainError, exception):
                domain.validate(candidate)
        else:
            self.assertEqual(domain.validate(candidate), exception)

    @parameterized.expand(
        [
            (
                NumpyStringDomain(allow_null=True),
                NumpyStringDomain(allow_null=True),
                True,
            ),
            (
                NumpyStringDomain(allow_null=True),
                NumpyStringDomain(allow_null=False),
                False,
            ),
            (
                NumpyStringDomain(allow_null=False),
                NumpyStringDomain(allow_null=False),
                True,
            ),
        ]
    )
    def test_eq(self, domain1: Domain, domain2: Domain, expected: bool):
        """Tests that __eq__ works correctly."""
        self.assertEqual(domain1 == domain2, expected)
